TYPE_CONV_PATH "Goba"

open VecOps
open GobaUtils
open GobaOps

let startup_lasers = 5

let base_attack_sound = Sfx.Sample "base-attack"

let recharge_distance = 50.0

let recharge_sound = Sfx.Sample "recharge"

let channel = Sfx.alloc_channel ()

let max_hp = 300.0
let max_laser_load = 10.0
let laser_accu_speed = 0.1

type init =
    { init_color	: (int * int * int);
      init_at		: float * float;
      init_hp		: float;
      init_laser_load	: float;
      init_bombs	: int;
      init_distance	: float;
      init_angle	: float;
    }
with sexp

let defaults =
  { init_color	= (255, 255, 255);
    init_at	= (0.0, 0.0);
    init_hp	= max_hp;
    init_laser_load = 0.0;
    init_bombs	= 4;
    init_distance = 150.0;
    init_angle  = 90.0 }

let base_bitmap_offset = (-60, -60)

let health_frames prev_hp hp color =
  let health_steps =
    [(0.0, "base_broken_color-0003", "base_broken_value-0003");
     (0.2, "base_broken_color-0002", "base_broken_value-0002");
     (0.4, "base_broken_color-0001", "base_broken_value-0001");
     (0.6, "base_broken_color-0000", "base_broken_value-0000");
     (0.8, "base_color", "base_value")]
  in
  let hp_frame hp =
    let n =
      snd (
	List.fold_left
	  (fun (n, frame) (step, color, value) ->
	     let n' = n + 1 in
	     if hp >= step then
	       (n', Some n)
	     else
	       (n', frame))
	  (0, None)
	  health_steps
      )
    in
    let n = GobaUtils.deoption n in
    let (_, color_bm, value_bm) = List.nth health_steps n in
    let nc = Bitmaps.no_cache in
      n, lazy (Bitmaps.bitmap @@ 
		 Bitmaps.move base_bitmap_offset @@
		 Bitmaps.compose 
		 [Bitmaps.colored color @@ Bitmaps.auto_crop @@ nc @@ Bitmaps.file value_bm;
		  Bitmaps.auto_crop @@ nc @@ Bitmaps.file color_bm
		 ])
  in
  let frame'old = hp_frame prev_hp in
  let frame'cur = hp_frame hp in
    snd frame'cur, fst frame'old <> fst frame'cur

class t id init0 =
  let init = Object.default_with init0 init_of_sexp in
  let color = init.init_color in
  let cmask = 
    let module B = Bitmaps in
      lazy 
	(B.mask_with_offset B.collision_mask_of_surface2 
	   (B.bitmap (B.move base_bitmap_offset (B.auto_crop (B.file "base-mask"))))) 
  in

object (self : 'a)
  inherit Object.t id as super
  inherit Collision.t

  val mutable at = init.init_at
  val mutable hp = init.init_hp
  val mutable next_recharge = 0.0
  val mutable under_attack_sound = false
  val mutable laser_load = init.init_laser_load
  val mutable laser_enabled = false
  val mutable bombs = init.init_bombs
  val mutable bitmap = fst (health_frames (init.init_hp /. max_hp) (init.init_hp /. max_hp) init.init_color)
  val mutable bitmap_changed = false
  val mutable angle = init.init_angle
  val mutable distance = init.init_distance

  method get_angle = angle

  method set_angle a = 
    angle <- a;
    self#change;

  method get_distance = distance
    
  method set_distance d = 
    distance <- d;
    self#change;

  method get_color = init.init_color

  method get_type = ObjectType.Base

  method get_laser_load =
    laser_load

  method get_bombs =
    bombs

  method get_state delay =
    Object.sexp_of_state sexp_of_init
      { Object.state_super = super#get_state delay;
	state_init = { init_color = color;
		       init_at = at;
		       init_hp = hp;
		       init_laser_load = laser_load;
		       init_bombs = bombs;
		       init_angle = angle;
		       init_distance = distance; } }

  method set_state state delay =
    let state = Object.state_of_sexp init_of_sexp state in
      super#set_state state.Object.state_super delay;
      let init = state.Object.state_init in
	hp <- init.init_hp;
	laser_load <- init.init_laser_load;
	bombs <- init.init_bombs;
	angle <- init.init_angle;
	distance <- init.init_distance;

  method set_location at' = at <- at'

  method location = at
  method set_speed speed' = ()
  method speed = 0.0, 0.0

  method z_location = 0.0

  method render gfx (_:Object.render_flags) at = 
    [(0, fun () -> Bitmaps.blit gfx at (Lazy.force bitmap))]

  method do_recharge (world:Object.world) =
    if self#self_owned then
      (*let at0 = at +|. Vector.float2 bitmap_offset in*)
      let at0 = at in
      let objs = 
	world#objs 
	  (at0 -|. Vector.v2 recharge_distance, at0 +|. Vector.v2 recharge_distance)
	  (fun o -> o#self_owned && Vector.abs2 (o#location -|. at0) < recharge_distance)
      in
	if List.fold_left (fun recharged o -> o#recharge || recharged) false objs then
	  if not (world#channel channel)#playing then
	    (world#channel channel)#play recharge_sound

  method tick delta (world:Object.world) = 
    if under_attack_sound then
      begin
	under_attack_sound <- false;
	if not (world#channel channel)#playing then
	  (world#channel channel)#play ~loops:(-1) base_attack_sound;
	(world#channel channel)#expire (Some 2.0);
      end;
    if laser_enabled then
      laser_load <- max 0.0 (laser_load -. delta)
    else
      laser_load <- min max_laser_load (laser_load +. laser_accu_speed *. delta);
    next_recharge <- next_recharge -. delta;
    if bitmap_changed then
      begin
	bitmap_changed <- false;
	world#explode ~sound:true ~kind:ExplosionKind.Massive (at +|. (Random.float 15.0 -. 7.0, Random.float 15.0 -. 7.0)) (0.0, 0.0);
      end;
    if next_recharge <= 0.0 then
      begin
	next_recharge <- 0.2;
	self#do_recharge world;
      end

  method move f = 
    at <- f at

  method is_alive = hp > 0.0

  method collision_area = 
    Collision.carea_of_mask at (Lazy.force cmask)

  method get_hp = hp

  method signal_bomb =
    if bombs > 0 then
      begin
	bombs <- bombs - 1;
	self#change;
	true;
      end
    else
      false

  method signal_laser state =
    if laser_load > 0.1 && state then
      begin
	if not laser_enabled then
	  begin
	    laser_enabled <- state;
	    self#change;
	  end;
	true
      end
    else
      begin
	if laser_enabled then
	  begin
	    laser_enabled <- false;
	    self#change;
	  end;
	false
      end

  method cause_damage damage =
    under_attack_sound <- true;
(*    if not (world.Object.world.world_channel channel)#playing then
    (*	    Sdlmixer.play_channel ~channel:7 (Lazy.force recharge_sound)

      Sdlmixer.play_channel ~loops:~-1 ~channel:7 (Lazy.force base_attack_sound);*)
(*    Sdlmixer.expire_channel 7 (Some 2.0);*)
    (world.Object.world.world_channel channel)#expire (Some 2.0);*)
    let hp' = hp in
      hp <- max 0.0 (hp -. damage);
      let bitmap', changed = health_frames (hp' /. max_hp) (hp /. max_hp) init.init_color in
	bitmap <- bitmap';
	bitmap_changed <- bitmap_changed || changed;
	self#change

  initializer
    (match init0 with 
       | Object.Init _ -> ()
       | Object.State (state0, age) -> self#set_state state0 age);
    self#change;
end
