open VecOps

type idx =
  [ `Char of (char * Sdlvideo.color)
  | `Block of (int * int * Sdlvideo.color * int)
  | `Bitmap of (string)
  | `Colored of (idx * Sdlvideo.color)
  | `Compose of (idx list)
  | `Move of ((int * int) * idx)
  | `Realpha of (float * idx)
  | `Brightness of (float * idx)
  | `Combine of (idx * idx * idx)
  | `Empty
  | `TileHoriz of (idx * int)
  | `TileVert of (idx * int)
  | `MirrorHoriz of idx
  | `MirrorVert of idx
  | `NoCache of idx
  | `Surface of (Sdlvideo.surface)
  | `Crop of (idx * ((int * int) * (int * int)))
  | `AutoCrop of idx
  | `Line of (int * int)
  | `Native of idx
      
  (* dims, function *)
  | `Function of ((int * int) * ((int * int) -> ((int * int * int) * int)))
  ]

type use_index = Int64.t

let garbage_interval = 100

type t = { surfaces : (idx, ((int * int) * Sdlvideo.surface) * (use_index ref)) Hashtbl.t;
	   cur_index : use_index ref;
	   garbage_counter : int ref }
    
let archive =
  lazy (
    GobaUtils.find_some 
      (fun path ->
	 let file = path ^ "/../gfx.tar" in
	   if try Unix.access file [Unix.R_OK]; true with Unix.Unix_error _ -> false then
	     Some (TarArchive.open_archive file)
	   else
	     None
      )
      Config.gfx_paths
  )

let char p : idx = `Char p
let block p : idx = `Block p
let file p : idx = `Bitmap p
let colored color idx = `Colored (idx, color)
let compose p : idx = `Compose p
let move ofs idx : idx = `Move (ofs, idx)
let realpha alpha idx : idx = `Realpha (alpha, idx)
let combine p : idx = `Combine p
let empty : idx = `Empty
let tile_horiz p : idx = `TileHoriz p
let tile_vert p : idx = `TileVert p
let mirror_horiz p : idx = `MirrorHoriz p
let mirror_vert p : idx = `MirrorVert p
let no_cache p : idx = `NoCache p
let crop area idx = `Crop (idx, area)
let auto_crop idx = `AutoCrop idx
let brightness v idx = `Brightness (v, idx)
let line p = `Line p
let native n = `Native n

let surface p = `Surface p
let surface_ofs (ofs, idx) = `Move (ofs, (`Surface idx))

let of_function p : idx = `Function p

let font = lazy (Gfx.font 25)

(* doesn't handle exceptions from lock/unlock-functions *)
let with_locked_surfaces s f =
  List.iter Sdlvideo.lock s;
  let r = GobaUtils.valuefy f () in
    List.iter Sdlvideo.unlock s;
    GobaUtils.pass r

let new_surface ?like_surface (w, h) =
  match like_surface with
    | None ->
	Sdlvideo.create_RGB_surface [`SRCALPHA] ~w ~h ~bpp:32 ~rmask:0xff0000l ~gmask:0x00ff00l ~bmask:0x0000ffl ~amask:0xff000000l
    | Some s ->
	Sdlvideo.create_RGB_surface_format s [`SRCALPHA] ~w ~h

let get_face (ch, color) =
  let font_height = Sdlttf.font_height (Lazy.force font) in
  let s = Sdlttf.render_glyph_blended (Lazy.force font) ch color in
  let (minx, _maxx, _miny, maxy) = Sdlttf.glyph_metrics (Lazy.force font) ch in
    ((minx, font_height - maxy), s)

let create () =
  { surfaces = Hashtbl.create 100;
    cur_index = ref 0L;
    garbage_counter = ref 0; }

let bitmaps = lazy (create ())

let get_bitmaps () = Lazy.force bitmaps

let collision_mask_of_surface s =
  let si = Sdlvideo.surface_info s in
  let module C = CollisionMask in
  let mask = C.create_mask si.Sdlvideo.w si.Sdlvideo.h in
  let set = C.set_mask mask in
    ( for y = 0 to si.Sdlvideo.h - 1 do
	  ( for x = 0 to si.Sdlvideo.w - 1 do
		set (x, y) (Sdlvideo.get_pixel s x y <> 0l)
	    done )
      done );
    mask

let collision_mask_of_surface2 s =
  let si = Sdlvideo.surface_info s in
  let module C = CollisionMask in
  let mask = C.create_mask si.Sdlvideo.w si.Sdlvideo.h in
  let { Sdlvideo.amask = alpha_mask } = Sdlvideo.surface_format s in
  let consider_mask = Int32.lognot alpha_mask in
  let set = C.set_mask mask in
    ( for y = 0 to si.Sdlvideo.h - 1 do
	  ( for x = 0 to si.Sdlvideo.w - 1 do
		set (x, y) (Int32.logand (Sdlvideo.get_pixel s x y) consider_mask = 0l)
	    done )
      done );
    mask

let mask_with_offset : (Sdlvideo.surface -> CollisionMask.mask) -> ((int * int) * Sdlvideo.surface) -> ((int * int) * CollisionMask.mask) =
  fun f (ofs, s) ->
    (ofs, f s)

let map_surface f src =
  let w, h = Gfx.surface_dims src in
  let dst = new_surface ~like_surface:src (w, h) in
    with_locked_surfaces [src; dst]
      (fun () ->
	 GobaUtils.iter2d 
	   (fun (x, y) ->
	      let ((r, g, b), alpha) = Sdlvideo.get_RGBA src (Sdlvideo.get_pixel src x y) in
		Sdlvideo.put_pixel dst x y 
		  (let (r, g, b, alpha) = 
		     f (r, g, b, alpha) in
		     Sdlvideo.map_RGB dst ~alpha (r, g, b)
		  )
	   )
	   ((0, 0), (w, h));
	 dst
      )

let init_surface_pixels ?like_surface (w, h) f =
  let dst = new_surface ?like_surface (w, h) in 
    with_locked_surfaces [dst]
      (fun () ->
	 GobaUtils.iter2d 
	   (fun (x, y) -> Sdlvideo.put_pixel dst x y (f (x, y)))
	   ((0, 0), (w, h));
	 dst
      )

let init_surface ?like_surface (w, h) f =
  let dst = new_surface ?like_surface (w, h) in
    with_locked_surfaces [dst]
      (fun () ->
	 GobaUtils.iter2d 
	   (fun (x, y) -> 
	      let (color, alpha) = f (x, y) in   
		Sdlvideo.put_pixel dst x y (Sdlvideo.map_RGB ~alpha dst color))
	   ((0, 0), (w, h));
	 dst
      )

(* oh why isn't this in the standard library.. *)
let array_for_all f a =
  let len = Array.length a in
  let rec aux idx =
    if idx >= len then
      true
    else
      if f a.(idx) then
	aux (idx + 1)
      else
	false
  in
    aux 0

let map_surfaces f srcs =
  let w, h = Gfx.surface_dims srcs.(0) in
    assert (array_for_all ((=) (w, h)) (Array.map Gfx.surface_dims srcs));
    let srcs_len = Array.length srcs in
    let dst = new_surface ~like_surface:srcs.(0) (w, h) in
    let data = Array.make srcs_len ((0, 0, 0), 0) in
      with_locked_surfaces (dst::Array.to_list srcs) 
	(fun () ->
	   GobaUtils.iter2d 
	     (fun ((x, y) as at) ->
		for c = 0 to srcs_len - 1 do
		  data.(c) <- Sdlvideo.get_RGBA srcs.(c) (Sdlvideo.get_pixel srcs.(c) x y)
		done;
		Sdlvideo.put_pixel dst x y
		  (let (color, alpha) = f at data in
		     Sdlvideo.map_RGB dst ~alpha color)
	     )
	     ((0, 0), (w, h));
	   dst
	)

(* attempts to ignore color 2 brightness *)
let color_near (r1, g1, b1) (r2, b2, g2) =
  let ar, ab, ag = float r1, float g1, float b1 in
  let cr, cg, cb = 
    float r2 /. ar,
    float g2 /. ag,
    float b2 /. ab
  in
  let min = min cr (min cg cb) in
  let max = max cr (max cg cb) in
  let delta = max -. min in
    delta /. max < (0.002 *. (ar +. ag +. ab) /. 3.0)

(* recolors white/gray *)
let recolor (dr, dg, db) ((r, g, b, a) as orig) =
  if color_near (0xa3, 0xa0, 0x87) (r, g, b) ||
    color_near (0x85, 0x81, 0x61) (r, g, b) then
      let br = float (List.fold_left (+) r [g; b]) /. (3.0 *. 255.0) in
      let color = (int_of_float (float dr *. br), 
		   int_of_float (float dg *. br),
		   int_of_float (float db *. br),
		   a)
      in
	color
  else
    orig

(* recolors all colors *)
let recolor_any (dr, dg, db) (r, g, b, a) =
  let br = float (List.fold_left (+) r [g; b]) /. (3.0 *. 255.0) in
  let color = (int_of_float (float dr *. br), 
	       int_of_float (float dg *. br),
	       int_of_float (float db *. br),
	       a) in
    color

let do_realpha adjust (r, g, b, a) =
  (r, g, b, min 255 (max 0 (adjust a)))

let do_brightness adjust (r, g, b, a) =
  let adj v = min 255 (max 0 (adjust v)) in
    (adj r, adj g, adj b, a)

let ipolate_rgba v ((r1, g1, b1), a1) ((r2, g2, b2), a2) =
  let v' = 1.0 -. v in
  let ( *! ) r v = int_of_float (r *. float v) in
    ((v' *! r1 + v *! r2,
      v' *! g1 + v *! g2, 
      v' *! b1 + v *! b2), 
     v' *! a1 + v *! a2)

let get surface x y = Sdlvideo.get_RGBA surface (Sdlvideo.get_pixel surface x y)

(*
let dump_surface s =
  let w, h = Gfx.surface_dims s in
    for y = 0 to h - 1 do
      for x = 0 to w - 1 do
	let color, alpha = get src xc yc in
      done;
      Printf.printf "\n"
    done
*)

let blit (gfx:#Gfx.t) at (ofs, surface) =
  gfx#blit (at +| ofs) surface

(* finally had to resort to: http://en.wikipedia.org/wiki/Alpha_transparency *)
let alpha_blit ~src ~dst ~dstofs =
  let s'w, s'h = Gfx.surface_dims src in
    with_locked_surfaces [src; dst]
      (fun () -> 
	 for xc = 0 to s'w - 1 do
	   for yc = 0 to s'h - 1 do
	     let dx, dy = xc + fst dstofs, yc + snd dstofs in
	     let src_color, src_alpha = get src xc yc in
	     let dst_color, dst_alpha = get dst dx dy in
	     let f255 v = float v /. 255.0 in
	     let new_alpha = int_of_float (min 255.0 (float src_alpha +. float dst_alpha *. (1. -. f255 src_alpha))) in 
	       if new_alpha > 1 then
		 let new_color = 
		   Vector.int3 (Vector.float3 src_color *||. Vector.v3 (float src_alpha /. float new_alpha) +||.
				  Vector.float3 dst_color *||. Vector.v3 ((float dst_alpha *. (1.0 -. f255 src_alpha)) /. float new_alpha)) in
		   Sdlvideo.put_pixel dst dx dy (Sdlvideo.map_RGB dst ~alpha:new_alpha new_color);
		   (*	      (let r's, g's, b's = src_color in
			      let r'd, g'd, b'd = dst_color in
			      let r, g, b = new_color in
			      Printf.printf "((%d,%d,%d),%d)+((%d,%d,%d),%d)->((%d,%d,%d),%d) " 
			      r's g's b's src_alpha
			      r'd g'd b'd dst_alpha
			      r g b new_alpha)*)
	       else
		 Sdlvideo.put_pixel dst dx dy (Sdlvideo.map_RGB dst ~alpha:0 dst_color);
	   done;
	   (*      Printf.printf "\n";*)
	 done
      )

let bm_region (ofs, bm) =
  let dims = Gfx.surface_dims bm in
    (ofs, ofs +| dims)

let move_region (a, b) ofs =
  (a +| ofs, b +| ofs)

let garbage_collect bm =
  let limit = Int64.sub !(bm.cur_index) 1000L in
  let removes = 
    Hashtbl.fold
      (fun key (_, r) l ->
	 if !r < limit then
	   key::l
	 else
	   l
      )
      bm.surfaces
      []
  in
    List.iter
      (Hashtbl.remove bm.surfaces)
      removes
    (*;
    if removes <> [] then
      Printf.printf "Removed %d surfaces\n%!" (List.length removes)*)
      

let rec bitmap ?(cache=true) (idx : idx) =
  let bm = Lazy.force bitmaps in
    match idx with
      | `Surface (surface) ->
	  ((0,0), surface)
      | `NoCache idx -> 
	  bitmap ~cache:false idx
      | `Function p ->
	  from_function p
      | other ->
	  try 
	    let (image, r) = Hashtbl.find bm.surfaces idx in
	      r := !(bm.cur_index);
	      image
	  with Not_found ->
	    let (ofs, surface) = 
	      match other with
		| `Char el -> get_face el
		| `Block (width, height, color, alpha) ->
		    (0, 0), 
		    (let s = new_surface (width, height) in
		       Sdlvideo.fill_rect s (Sdlvideo.map_RGB ~alpha s color);
		       s)
		| `Bitmap (file) ->
		    let name path ext = Printf.sprintf "%s/%s.%s" path file ext in
		    let exts = ["png"; "jpg"] in
		      ((0, 0),
		       (GobaUtils.find_some (
			  fun (path, ext) ->
			    try 
			      (* sdl apparently doesn't like when we open non-existing files.. *)
			      match path with
				| "%ARCHIVE%" ->
				    let file = file ^ "." ^ ext in
				      ( match GobaUtils.valuefy (TarArchive.find (Lazy.force archive)) file with
					  | `Value v -> 
					      (*let _ = Printf.printf "Loading %s from archive\n%!" file in*)
						Some (Sdlloader.load_image_from_mem (TarArchive.string_of_contents v))
					  | `Exn e ->
					      None
				      )
				| path ->
				    let file = name path ext in
				      if try Unix.access file [Unix.R_OK]; true with Unix.Unix_error _ -> false then
					(*let _ = Printf.printf "Loading %s\n%!" file in*)
					Some (Sdlloader.load_image file)
				      else
					None
			    with Sdlloader.SDLloader_exception _ -> None)
			  (GobaUtils.list_product ("%ARCHIVE%"::Config.gfx_paths) exts)))
		| `Colored (idx, color) ->
		    let bm = bitmap idx in
		      (fst bm, map_surface (recolor_any color) (snd bm))
		| `Compose (idxs) -> do_compose idxs
		| `Combine (idxs) -> do_combine idxs
		| `Move (ofs, idx) ->
		    let bm = bitmap idx in
		      (fst bm +| ofs, snd bm)
		| `Realpha (adjust, idx) ->
		    let bm = bitmap idx in
		      (fst bm, map_surface (do_realpha (fun a -> int_of_float (float a *. adjust))) (snd bm))
		| `Brightness (adjust, idx) ->
		    let bm = bitmap idx in
		      (fst bm, map_surface (do_brightness (fun a -> int_of_float (float a *. adjust))) (snd bm))
		| `Empty ->
		    (0, 0), Sdlvideo.create_RGB_surface [] ~w:0 ~h:0 ~bpp:24 ~rmask:0xff0000l ~gmask:0x00ff00l ~bmask:0x0000ffl ~amask:0x0l
		| `TileHoriz p -> do_tile_horiz p
		| `TileVert p -> do_tile_vert p
		| `MirrorHoriz p -> do_mirror_horiz p
		| `MirrorVert p -> do_mirror_vert p
		| `Crop p -> do_crop p
		| `AutoCrop p -> do_auto_crop p
		| `NoCache _ | `Surface _ | `Function _ -> failwith "argh"
		| `Line p -> do_line p
		| `Native n -> 
		    let (ofs, surface) = bitmap n in
		      (ofs, Sdlvideo.display_format ~alpha:true surface)
	    in
	      bm.cur_index := Int64.add !(bm.cur_index) 1L;
	      incr bm.garbage_counter;
	      if !(bm.garbage_counter) > garbage_interval then
		begin
		  bm.garbage_counter := 0; 
		  garbage_collect bm;
		end;
	      let s = (ofs, surface) in
		if cache then
		  begin
		    Hashtbl.add bm.surfaces idx (s, ref !(bm.cur_index));
		  end;
		s

and do_compose idxs =
  if idxs = [] then
    bitmap `Empty
  else
    let bms = List.map bitmap idxs in
    let extreme f cmp = List.fold_left (fun v b -> if cmp v (f b) then v else (f b)) (f (List.hd bms)) (List.tl bms) in
    let min_x = extreme (fun ((x0, _), _) -> x0) (<) in
    let min_y = extreme (fun ((_, y0), _) -> y0) (<) in
    let max_x = extreme (fun ((x0, _), s) -> x0 + fst (Gfx.surface_dims s)) (>) in
    let max_y = extreme (fun ((_, y0), s) -> y0 + snd (Gfx.surface_dims s)) (>) in
    let ofs = min_x, min_y in
    let size = (max_x, max_y) -| (min_x, min_y) in
    let surface = new_surface size in
      Sdlvideo.fill_rect surface 0x00000000l;
      (*    Printf.printf "Begin composing\n";*)
      List.iter 
	(fun (ofs', bm) ->
	   let ofs = ofs' -| ofs in
	     (*	 let size = Gfx.surface_dims bm in*)
	     (*	   Sdlvideo.blit_surface 
		   ~src:bm
		   ~dst:surface
		   ~dst_rect:(Sdlvideo.rect (fst ofs) (snd ofs) (fst size) (snd size))
		   ()*)
	     alpha_blit ~src:bm ~dst:surface ~dstofs:ofs
	)
	bms;
      (*    Printf.printf "Finish composing\n%!";*)
      ofs, surface

and do_combine (value, left, right) =
  let idxs = [|value; left; right|] in
  let bms = Array.map bitmap idxs in
  let f _ vs =
    let v = 
      let ((r, g, b), _) = vs.(0) in
	float (r + g + b) /. (255.0 *. 3.0)
    in
      ipolate_rgba v vs.(1) vs.(2)
  in
    assert (array_for_all (fun (ofs, _) -> ofs = fst bms.(0)) bms);
    (fst bms.(0), 
     map_surfaces f (Array.map snd bms))

and do_tile_horiz (idx, p) = 
  let ((_, y'ofs), surface) = bitmap idx in
  let dims = Gfx.surface_dims surface in
    ((0, y'ofs), 
     init_surface_pixels ~like_surface:surface (fst dims * p, snd dims)
       (fun (x, y) -> Sdlvideo.get_pixel surface (x mod fst dims) y))

and do_tile_vert (idx, p) = 
  let ((x'ofs, _), surface) = bitmap idx in
  let dims = Gfx.surface_dims surface in
    ((x'ofs, 0), 
     init_surface_pixels ~like_surface:surface (fst dims, snd dims * p)
       (fun (x, y) -> Sdlvideo.get_pixel surface x (y mod snd dims)))

and do_mirror_horiz idx =
  let ((x'ofs, y'ofs), surface) = bitmap idx in
  let (x'size, y'size) = Gfx.surface_dims surface in
    ((-x'ofs - x'size, y'ofs),
     init_surface_pixels ~like_surface:surface (x'size, y'size)
       (fun (x, y) -> Sdlvideo.get_pixel surface (x'size - 1 - x) y))

and do_mirror_vert idx =
  let ((x'ofs, y'ofs), surface) = bitmap idx in
  let (x'size, y'size) = Gfx.surface_dims surface in
    ((x'ofs, -y'ofs - y'size),
     init_surface_pixels ~like_surface:surface (x'size, y'size)
      (fun (x, y) -> Sdlvideo.get_pixel surface x (y'size - 1 - y)))

and from_function (dims, f) =
  ((0, 0), init_surface dims f)

and do_auto_crop idx =
  let (ofs, src) = bitmap idx in 
  let (w, h) = Gfx.surface_dims src in
  let mask =
    let { Sdlvideo.amask = alpha_mask } = Sdlvideo.surface_format src in
      alpha_mask
  in
  let want_pixel (x, y) = 
    let v = (Sdlvideo.get_pixel src x y) in
    let mv = Int32.logand v mask in
      if mv <> 0l then
	Some ()
      else
	None
  in
  let where ?(label="") at0 dir n f =
    let rec aux at n' =
      if n' < n then
	match f at with
	  | None -> aux (at +| dir) (n' + 1)
	  | Some x -> 
	      (*Printf.printf "%s At %d\n" label n;*)
	      Some (n', x)
      else
	None
    in
      aux at0 0
  in
  let want_any_pixel at0 dir n = where ~label:"pixel" at0 dir n want_pixel in
  let top, bottom, left, right =
    with_locked_surfaces [src]
      (fun () ->
	 let scan at0 dir1 n1 dir2 n2 =
	   fst (
	     GobaUtils.option_default
	       (0, (0, ()))
	       (where ~label:"scan" at0 dir1 n1
		  (fun at -> want_any_pixel at dir2 n2))
	   )
	 in
	   let top = scan (0, 0) (0, 1) h (1, 0) w in
           let bottom = scan (0, h - 1) (0, -1) h (1, 0) w in
	   let left = scan (0, 0) (1, 0) w (0, 1) h in
	   let right = scan (w - 1, 0) (-1, 0) w (0, 1) h in
	     top, bottom, left, right
      )
  in
    let (x1, y1), (x2, y2) = (left, top), (w - right, h - bottom) in
      if x2 <= x1 || y2 <= y1 then
	bitmap `Empty
      else
	(* skips caching.. *)
	do_crop (idx, ((x1, y1) +| ofs, (x2, y2) +| ofs))

and do_crop (idx, ((x1, y1), (x2, y2))) =
  let bm = bitmap idx in
  let (ofs, src) = bm in
  let dims = Gfx.surface_dims src in
  let within_src (x, y) = x >= 0 && y >= 0 && x < fst dims && y < snd dims in

  let src'a = (x1, y1) -| ofs in
  let w = x2 - x1 in
  let h = y2 - y1 in

    with_locked_surfaces [src]
      (fun () ->
	 (ofs +| src'a,
	  init_surface_pixels ~like_surface:src (w, h)
	    (fun at -> 
	       let at' = at +| src'a in
		 if within_src at' then
		   let (x, y) = at' in Sdlvideo.get_pixel src x y
		 else
		   0l
	    )
	 )
      )

(* FIXME: broken *)
and do_line at =
  let cax = Vector.coaxel (Vector.float2 at) in
    ((0, 0),
     init_surface at
       (fun at' ->
	  let v = max 0.0 (1.0 -. 2.0 *. abs_float (Vector.dot (Vector.float2 at') cax)) in
	  let c = if v > 0.0 then (255, 255, 255) else (0, 0, 0) in
	    (c, int_of_float (255.0 *. v)) 
       )
    )

let format_size s =
  let rec aux s =
    if s >= 1000 then
      aux (s / 1000) ^ "'" ^ Printf.sprintf "%03d" (s mod 1000)
    else
      string_of_int s
  in
    aux s

let blit_idx (gfx:#Gfx.t) at idx =
  let (ofs, surface) = bitmap (`Native idx) in
    gfx#blit (at +| ofs) surface

let dump_stats () =
  let bms = Lazy.force bitmaps in
  let num_surfaces = Hashtbl.length bms.surfaces in
  let size = 
    Hashtbl.fold
      (fun _ ((_, surface), _) pixels ->
	     let si = Sdlvideo.surface_info surface in
	     let sf = Sdlvideo.surface_format surface in
	       pixels + sf.Sdlvideo.bytes_pp * si.Sdlvideo.w * si.Sdlvideo.h)
	  bms.surfaces
	  0
  in
  Printf.printf "Bitmaps stats: %d surfaces, %s bytes, %s bytes per surface\n" 
    num_surfaces
    (format_size size)
    (if num_surfaces > 0 then format_size (size / num_surfaces) else "N/A")

let _ = at_exit dump_stats
