(* Different kinds of buildings here, because they all 
   practically do the same thing.. exist. *)

TYPE_CONV_PATH "Goba"

open VecOps
open GobaOps

let bitmap_offset = (-60, -60)

type kind =
  | FakeBase of (int * int * int)	(* color *)
  | Wall of float			(* facing angle *)
with sexp

type init = { init_at		: float * float;
	      init_hp		: float;
	      init_kind		: kind }
with sexp

let fake_base_max_hp = 30.0

let kind_hp = function
  | FakeBase _ -> fake_base_max_hp
  | Wall _ -> 50.0

let defaults = { init_at = (0.0, 0.0);
		 init_hp = fake_base_max_hp;
		 init_kind = FakeBase (255, 255, 255)
	       }

class t id init0 =
  let init = Object.default_with init0 init_of_sexp in
object (self : 'a)
  inherit Object.t id as super

  val mutable hp = init.init_hp
  val mutable alive = true
  val mutable bitmap = 
    match init.init_kind with
      | FakeBase color ->
	  let max_hp = kind_hp init.init_kind in
	  fst (Base.health_frames (init.init_hp /. max_hp) (init.init_hp /. max_hp) color)
      | Wall angle ->
	  lazy (Bitmaps.bitmap @@ Bitmaps.move (-60, -60) @@ Bitmaps.auto_crop @@ Bitmaps.file @@ Printf.sprintf "wall-%04d" (1 + GobaMath.frame_of_radian angle))
  val mutable bitmap_changed = false
  val cmask = 
    let module B = Bitmaps in
      lazy 
	(B.mask_with_offset B.collision_mask_of_surface2 
	 @@ B.bitmap 
	 @@ B.move bitmap_offset 
	 @@ B.auto_crop 
	 @@ B.file
	 @@ match init.init_kind with
	   | FakeBase _ -> "base-mask"
	   | Wall angle -> Printf.sprintf "wall_mask-%04d" (1 + GobaMath.frame_of_radian angle)
	)

  method get_state delay =
    Object.sexp_of_state sexp_of_init
      { Object.state_super = super#get_state delay;
	state_init = { init_at = init.init_at;
		       init_hp = hp;
		       init_kind = init.init_kind;
		     } }

  method set_state state delay =
    let state = Object.state_of_sexp init_of_sexp state in
      super#set_state state.Object.state_super delay;
      let init = state.Object.state_init in
	hp <- init.init_hp

  method set_speed _ = ()
  method set_location _ = ()

  method tick delta world =
    let died = 
      if hp <= 0.0 && alive then
	begin
	  alive <- false;
	  self#change;
	  true
	end
      else
	false
    in
      if bitmap_changed || died then
	begin
	  bitmap_changed <- false;
	  world#explode ~sound:true ~kind:ExplosionKind.Massive (init.init_at +|. (Random.float 15.0 -. 7.0, Random.float 15.0 -. 7.0)) (0.0, 0.0);
	end;

  method location = init.init_at
  method speed = 0.0, 0.0

  method render gfx flags at =
    [(0, fun () -> Bitmaps.blit gfx at (Lazy.force bitmap))]

  method cause_damage damage =
    let hp' = hp in
      hp <- max 0.0 (hp -. damage);
      self#change;
      match init.init_kind with
	| FakeBase color ->
	    let max_hp = kind_hp init.init_kind in
	    let bitmap', changed = Base.health_frames (hp' /. max_hp) (hp /. max_hp) color in
	      bitmap <- bitmap';
	      bitmap_changed <- bitmap_changed || changed;
	| Wall _ ->
	    ()

  method is_alive = alive

  method get_type = ObjectType.Building

  method collision_area =
    Collision.carea_of_mask init.init_at (Lazy.force cmask)

  initializer
    (match init0 with 
       | Object.Init _ -> ()
       | Object.State (state0, age) -> self#set_state state0 age);
    self#change;
end

let has_space_for : Object.base -> (((float * float) * (float * float)) -> Object.base list) -> bool =
  fun obj objs ->
    let (region, _) = obj#collision_area in
      Collision.collisions obj (objs region) (GobaUtils.const true) = []
