TYPE_CONV_PATH "Goba"

open VecOps
open ExplosionKind
open GobaOps

type damage = float * float * float * float
with sexp

let default_damage : damage = (40.0, 4.0, 20.0, 2.0)

let damage_radius ((radius, _, _, _) : damage) = radius

type init = { init_at : float * float;
	      init_burst : float * float;
	      init_sound : bool;
	      init_damage : damage;
	      init_kind : ExplosionKind.t }
with sexp

let bitmap_offset = (-60, -60)

let frame_rate = 30.0

let bitmaps =
  Array.init 36
    (fun n ->
       lazy (
	 let module B = Bitmaps in
	 let alpha = GobaMath.clamp 0.0 1.0 (2.0 -. 3.0 *. float n /. 35.0) in
	 let bm = B.bitmap @@ B.realpha alpha (B.move bitmap_offset @@ B.auto_crop @@ B.no_cache @@ B.file @@ Printf.sprintf "explosion-%04d" (n + 1)) in
	   bm))

let explosion_sound = Sfx.Sample "hit"

let explosion_sound_massive = Sfx.Sample "explosion"

type pixel = 
    { mutable at : float * float;
      mutable speed : float * float;
      mutable life_ofs : float;
    }

let damage_function (radius, max_damage, p1, p2) d =
  max (0.0) (max_damage /. ((d /. p1 +. 1.0) ** p2) *. (1.0 -. d /. radius))

let explosion_pixels = 20

let explosion_color =
  let red = Vector.float3 (255, 0, 0) in
  let yellow = Vector.float3 (255, 255, 128) in
    fun brightness ->
      (* quantize the value, so we won't generate too many bitmaps (which will be cached) *)
      let r = float (Random.int 10) /. 10.0 in
      let r' = 1.0 -. r in
	Vector.int3 (red *||. Vector.v3 (brightness *. r) +||. yellow *||. Vector.v3 (brightness *. r'))

let channel = Sfx.alloc_channel ()
let prev_sound = ref 0.0

class t id init0 =
  let init = ref (Object.default_with init0 init_of_sexp) in

object (self)
  inherit Object.t id as super

  val mutable at = (!init).init_at
  val mutable time_left = 1.0
  val mutable alive = true
  val mutable sound = (!init).init_sound
  val mutable frame = 0.0
  val mutable pixels = 
    Array.of_list (
      GobaUtils.foldl1d 
	(fun _ l ->
	   let ang = Random.float (4.0 *. atan 1.0 *. 2.0) in
	   let dist = 15.0 *. ((Random.float 1.0) ** 1.3) in
	   let ofs = (cos ang *. dist, sin ang *. dist) in
	   let life_ofs = ~-. (Random.float 0.3) in
	   let spd = (ofs +|. (!init).init_burst) *|. Vector.dup 1.5 in
	     { at = ofs;
	       speed = spd;
	       life_ofs = life_ofs }::l
	)
	[]
	(0, explosion_pixels)
    )
      
  val mutable damage_left = 1.0

  method get_type = ObjectType.Explosion

  method set_location at' = at <- at'
  method location = at
  method z_location = -.5000.0

  method get_state delay =
    Object.sexp_of_state 
      sexp_of_init
      { Object.state_super = super#get_state delay;
	state_init = !init }

  method set_state state delay =
    let state = Object.state_of_sexp init_of_sexp state in
      super#set_state state.Object.state_super delay

  method set_speed (_ : float * float) = ()
  method speed = 0.0, 0.0

  method render gfx _ display_at =
    [(0, fun () ->
	if self#is_alive then
	  begin
	    if (!init).init_kind = Massive then
	      begin
		let frame_num = int_of_float frame in
		  if frame_num < 36 then
		    Bitmaps.blit gfx display_at (Lazy.force bitmaps.(frame_num));
	      end;
	    Array.iter 
	      (fun { at = ofs; life_ofs = life_ofs } ->
		 let clamp low high v = min high (max low v) in
		 let left = life_ofs +. time_left in
		   gfx#rectangle ~alpha:(clamp 0 255 (int_of_float (255.0 *. left))) (explosion_color (clamp 0.0 1.0 left)) (display_at +| Vector.int2 ofs) (1, 2))
	      pixels
	  end)]

  method impact world (_:float) = ()

  method collidable = false

  method destructible = false

  method collision_area = 
    ((0.0, 0.0), (0.0, 0.0)), CollisionMask.create_mask 0 0

  method tick delta world =
    time_left <- time_left -. delta;
    frame <- frame +. delta *. frame_rate;
    if sound then
      begin
	sound <- false;
	let snd = 
	  match !(init).init_kind with
	    | ExplosionKind.Massive -> explosion_sound_massive
	    | ExplosionKind.Particle -> explosion_sound
	in
	let chan = (world#channel channel) in
	let now = Unix.gettimeofday () in
	  if now -. !prev_sound > 0.2 || not chan#playing then
	    begin
	      prev_sound := now;
	      chan#play snd;
	    end
      end;
    let new_alive = time_left > 0.0 in
      if alive <> new_alive then
	self#change;
      alive <- new_alive;
      (*      pixels <- 
	      List.map (fun (at, speed, life_ofs) ->
	      at +|. speed *|. Vector.dup delta, speed, life_ofs) pixels*)
      if alive then
	Array.iteri (fun i { at = at; speed = speed } ->
		       pixels.(i).at <- at +|. speed *|. Vector.dup delta;
		       pixels.(i).speed <- speed *|. Vector.dup (0.5 ** delta);
		    ) pixels;
      if alive && self#self_owned then
	let old_damage = damage_left in
	  damage_left <- damage_left -. delta;
	  let delta = int_of_float old_damage - int_of_float damage_left in
	  let radius = damage_radius (!init).init_damage in
	    if delta > 0 then
	      let area = ((at -|. Vector.dup radius,
			   at +|. Vector.dup radius)) in
	      let objs = 
		world#objs area (fun obj -> Vector.abs2 (obj#location -|. at) <= radius) in
		List.iter (fun o -> 
			     o#impact world (0.5 +. (damage_function (!init).init_damage (Vector.abs2 (o#location -|. at))))) objs

  method is_alive = alive

  initializer
    (match init0 with 
       | Object.Init _ -> ()
       | Object.State (state0, age) -> self#set_state state0 age);
    self#change;
end
