TYPE_CONV_PATH "Goba"
open VecOps
open GobaMath
open GobaOps
open GobaGfx

module B = Bitmaps

type tank_id = TankId of int

type 'a object_list = (Object.object_id, 'a) Hashtbl.t

let sync_interval = 1.0 /. 30.0

let tank_sound = Sfx.Sample "tankkiloop"

let degree = Char.chr 0xb0

type t = { player_id	: Player.id;
	   players	: (Player.id * Protocol.color * Object.object_id) list;
	   map		: GobaMap.t;
	   my_tanks	: (tank_id, Object.object_id) Hashtbl.t;
	   mutable base	: Object.object_id;
	   mutable laser : Object.object_id option;
	   client	: Client.t;

	   mutable pending_ping : bool;
	   ping_value	: float option ref;

	   object_old_map : ((int * int) list) object_list;
	   object_map	: Object.object_id list array;
	   object_removed : Object.object_id list ref;

	   objects	: Object.base object_list;
	   tanks	: Tank.t object_list;
	   bases	: Base.t object_list;

	   latency	: float ref;

	   mutable object_map_serial : Object.serial;
	   mutable last_sync : float;
	   mutable last_sync_serial : Object.serial;

	   mutable object_serial : int64;
	   object_increment	: int64;

	   mutable endgame_view	: float * float;
	   messages	: (string * Sdlvideo.surface) list ref;
	 }

let tank_channels = Array.init 4 (fun _ -> new SfxSdl.channel (Sfx.alloc_channel ()))
let fire_sound = Sfx.Sample "ampu"
let fire_channel = new SfxSdl.channel (Sfx.alloc_channel ())
let base_laser_sound = Sfx.Sample "base-laser"
let base_laser_channel = new SfxSdl.channel (Sfx.alloc_channel ())

exception JoinFailed

let symbol name = lazy (Bitmaps.bitmap @@ B.native @@ (Bitmaps.move (-60, -60) (Bitmaps.file (name ^ "_symbol"))))

let bomb_symbol = symbol "bomb"
let angle_symbol = symbol "angle"
let distance_symbol = symbol "distance"

let (fade_nw, fade_ne, fade_sw, fade_se, fade_horiz, fade_vert) = 
  let f (x, y) =
    let v = int_of_float (255.0 *. (max 0.0 (7.0 -. sqrt (float x ** 2.0 +. float y ** 2.0))) /. 7.0) in
      (255, 255, 255), v
  in
    (lazy (Bitmaps.bitmap @@ B.native @@ (B.mirror_vert (B.mirror_horiz (B.move (0, 0) (B.of_function ((8, 8), f)))))),
     lazy (Bitmaps.bitmap @@ B.native @@ (B.mirror_vert ((B.move (0, 0) (B.of_function ((8, 8), f)))))),
     lazy (Bitmaps.bitmap @@ B.native @@ ((B.mirror_horiz (B.move (0, 0) (B.of_function ((8, 8), f)))))),
     lazy (Bitmaps.bitmap @@ B.native @@ (((B.move (0, 0) (B.of_function ((8, 8), f)))))),
     lazy (Bitmaps.bitmap @@ B.native @@ (B.move (0, 0) (B.of_function ((8, 1), f)))),
     lazy (Bitmaps.bitmap @@ B.native @@ (B.move (0, 0) (B.of_function ((1, 8), f)))))

let fade_east l =
  Bitmaps.bitmap @@ B.native @@ (B.tile_vert (B.surface_ofs (Lazy.force fade_horiz), l))
    
let fade_west l =
  Bitmaps.bitmap @@ B.native @@ (B.tile_vert (B.mirror_horiz (B.surface_ofs (Lazy.force fade_horiz)), l))
    
let fade_north l =
  Bitmaps.bitmap @@ B.native @@ (B.tile_horiz (B.mirror_vert (B.surface_ofs (Lazy.force fade_vert)), l))
    
let fade_south l =
  Bitmaps.bitmap @@ B.native @@ (B.tile_horiz ((B.surface_ofs (Lazy.force fade_vert)), l))

let fade_rect gfx ((x1, y1), (w, h)) =
  let x2, y2 = x1 + w, y1 + h in
    Bitmaps.blit gfx (x1, y1) (Lazy.force fade_nw);
    Bitmaps.blit gfx (x2, y1) (Lazy.force fade_ne);
    Bitmaps.blit gfx (x1, y2) (Lazy.force fade_sw);
    Bitmaps.blit gfx (x2, y2) (Lazy.force fade_se);
    Bitmaps.blit gfx (x1, y1) (fade_west h);
    Bitmaps.blit gfx (x2, y1) (fade_east h);
    Bitmaps.blit gfx (x1, y1) (fade_north w);
    Bitmaps.blit gfx (x1, y2) (fade_south w)

let background =
  lazy (snd (Bitmaps.bitmap @@ B.native @@ (B.file "background")))
    
(* lot's of magic happens in this function!

better cut down the side effects of the protocol somehow.. 

start asynchronous receiving events after this initialization? needs
protocol changes.

*)
let create client player_num =
  let tanks = Hashtbl.create 10 in
  let objects = Hashtbl.create 10 in
  let bases = Hashtbl.create 10 in
  let messages = ref [] in
  let object_removed = ref [] in

  let latency = ref 0.0 in

  let handle_object { Protocol.oi_id = id;
		      oi_pid = owner;
		      oi_state = state;
		      oi_type = otype } delay =
    match id with
      | Object.MapId at -> failwith "Cannot handle map objects!"
      | Object.ObjectId id ->
	  let delay = delay +. !latency /. 2.0 in
	    if try ignore (Hashtbl.find objects id); false with Not_found -> true then
	      let register hash obj = Hashtbl.add hash (Object.to_object_id (obj#get_id)) obj in
	      let obj =
		match otype with
		  | ObjectType.Base ->
		      let obj = new Base.t id (Object.State (state, delay)) in
			register bases obj;
			(obj :> Object.base);
		  | ObjectType.Tank ->
		      let obj = new Tank.t id (Object.State (state, delay)) in
			register tanks obj;
			(obj :> Object.base);
		  | _ -> ObjectFactory.create otype (Object.ObjectId id) (state, delay)
	      in
		register objects obj;
		if owner = Player.Id player_num then
		  obj#set_self_owned true;
	    else
	      let obj = Hashtbl.find objects id in
		obj#set_state state delay
  in
  let handle_destruct id =
    object_removed := Object.to_object_id id :: !object_removed
  in
  let map = 
    lazy (let map = Client.map client in
	    GobaMap.create map) 
  in
  let handle_chat (from, msg) =
    let msg = from ^ "> " ^ msg in
      if msg <> "" then
	begin
	  let surface = Sdlttf.render_text_blended (Gfx.get_font ~size:20 ()) msg (200, 255, 128) in
	    messages := (msg, surface) :: !messages;
	    if List.length !messages > 5 then
	      messages :=
		fst (
		  GobaUtils.foldl1d
		    (fun _ (new_msgs, msgs) ->
		       List.hd msgs::new_msgs, List.tl msgs
		    )
		    ([], !messages)
		    (0, 5)
		)
	end
  in
  let handle_damage_object (id, damage) =
    match id with
      | Object.MapId at ->
	  let tile, { Tile.hp = hp } = GobaMap.at_full (Lazy.force map) at in
	    GobaMap.set 
	      (Lazy.force map) at (tile, { Tile.hp = hp -. damage })
      | Object.ObjectId id ->
	  match try Some (Hashtbl.find objects id) with Not_found -> None with
	    | Some obj ->
		(*Printf.printf "Causing damage to object %Ld\n%!" (Object.int_of_id id);*)
		obj#cause_damage damage
	    | None -> ()
  in
    Client.set_object_handler client handle_object;
    Client.set_destruct_handler client handle_destruct;
    Client.set_damage_object_handler client handle_damage_object;
    Client.set_chat_handler client handle_chat;
    let player_id, base_id, my_tanks, (obj_base, obj_increment) =
      let my_tanks = Hashtbl.create 4 in
      let player_id, obj_ids, obj_constr = 
	(*Printf.printf "Begin join\n%!";*)
	match Client.join client player_num (Unix.getlogin ()) with
	  | Some x -> x
	  | None -> raise JoinFailed
      in
      let rec wait_objs () =
	(*Printf.printf "Waiting %s\n%!" (String.concat " " (List.map (fun id -> Object.string_of_id (Object.ObjectId id)) obj_ids));*)
	Client.handle_io client;
	(try ignore(Unix.select [] [] [] 0.1) with exn -> ());
	if List.for_all (fun id -> Hashtbl.mem objects id) obj_ids then
	  ()
	else
	  wait_objs ()
      in
	(*Printf.printf "Begin wait\n%!";*)
	wait_objs ();
	(*Printf.printf "Wait done\n%!";*)
	ignore (List.fold_left 
		  (fun cur_id tank_id ->
		     let obj = Hashtbl.find objects tank_id in
		       obj#set_self_owned true;
		       match obj#get_type with
			 | ObjectType.Tank ->
			     Hashtbl.add my_tanks (TankId cur_id) tank_id;
			     (*let tank = Hashtbl.find tanks tank_id in
			       tank#set_ident (Printf.sprintf "%d" (cur_id + 1));*)
			       cur_id + 1
			 | ObjectType.Base ->
			     cur_id
			 | _ -> failwith "Types other than Tank and Base not supported as client objects"
		  ) 0 obj_ids);
	player_id, (List.hd obj_ids), my_tanks, obj_constr
    in
    let players = Client.list_players client in
    let map = Lazy.force map in
      { player_id	= player_id;
	players		= players;
	client		= client;
	map		= map;
	my_tanks	= my_tanks;
	base		= base_id;
	laser		= None;
	objects		= objects;
	tanks		= tanks;
	bases		= bases;
	last_sync	= 0.0;
	last_sync_serial = Object.no_serial;
	latency		= latency;
	object_serial	= obj_base;
	object_increment = obj_increment;
	pending_ping	= false;
	ping_value	= ref None;
	messages	= messages;

	object_map	= Array.create (map.GobaMap.width * map.GobaMap.height) [];
	object_old_map  = Hashtbl.create 100;
	object_map_serial = Object.no_serial;
	object_removed = object_removed;
	
	endgame_view	= (try (Hashtbl.find objects base_id)#location with Not_found -> 0.0, 0.0);
      }

let tank t id = 
  try Some (Hashtbl.find t.tanks (Hashtbl.find t.my_tanks id))
  with Not_found -> None

let base t = 
  try Some (Hashtbl.find t.bases t.base)
  with Not_found -> None

let find_object t obj =
  Hashtbl.find t.objects obj

let my_tanks t =
  GobaUtils.list_of_hash (fun _ oid -> (find_object t oid :> Object.base)) t.my_tanks

let objects ?except ?collidables ?self_owned ?changed_since ?filter t =
  Hashtbl.fold (fun id obj l -> 
		  if 
		    (collidables = None || Some obj#collidable = collidables)
		      && (self_owned = None || Some obj#self_owned = self_owned)
		      && (match changed_since with 
			    | None -> true
			    | Some t -> obj#changed_since t)
		      && (match except with 
			    | None -> true
			    | Some obj' -> obj != obj')
		      && (match filter with
			    | None -> true
			    | Some filter -> filter obj)
		  then
		    obj::l
		  else
		    l) t.objects []

let accel_tank t id adjust =
  match tank t id with
    | None -> ()
    | Some tank ->
	tank#accelerate adjust;
	let (TankId tank_num) = id in
	let chan = tank_channels.(tank_num) in
	  if not chan#playing then
	    chan#play ~loops:~-1 tank_sound;
	  chan#expire (Some 0.5)

let turn_tank t id dir =
  match tank t id with
    | None -> ()
    | Some tank ->
	tank#turn dir

let new_object_id t =
  let id = Object.object_id_of_int64 t.object_serial in
    t.object_serial <- Int64.add t.object_serial t.object_increment;
    id

let object_info t o =
  let id, at, speed = o#get_id, o#location, o#speed in
    id, at +|. speed *|. Vector.dup (!(t.latency) /. 2.0), speed

let new_object_info t o : Protocol.object_info =
  let id = o#get_id in
    { Protocol.oi_id = id;
      oi_pid = t.player_id;
      oi_state = o#get_state (!(t.latency) /. 2.0);
      oi_type = o#get_type }

let create_object t o =
  Client.create_object t.client (new_object_info t o);
  o#set_self_owned true;
  Hashtbl.add t.objects (Object.to_object_id o#get_id) (o :> Object.base) 

let within x (x1, x2) = x >= x1 && x <= x2

let region t (a, b) (expand_x, expand_y) = 
  let (x1, y1) = Vector.round2 (Vector.floor2 (a /|. Vector.float2 t.map.GobaMap.tile_size)) in
  let (x2, y2) = Vector.round2 (Vector.ceil2 (b /|. Vector.float2 t.map.GobaMap.tile_size)) in
  let x1 = max 0 (x1 - expand_x) in
  let y1 = max 0 (y1 - expand_y) in
  let x2 = min (t.map.GobaMap.width - 1) (x2 + expand_x) in
  let y2 = min (t.map.GobaMap.width - 1) (y2 + expand_y) in
  let (a, b) = ((x1, y1), (x2, y2)) in
    (a, b +| (1, 1))

let objects_within t region' =
  let region = region t region' (1, 1) in
  let w, h = t.map.GobaMap.width, t.map.GobaMap.height in
  let objs = List.concat (
    GobaUtils.foldl2d
      (fun (x, y) objs ->
	 t.object_map.(x + y * w)::objs
      )
      []
      region
  )
  in
  let ids = GobaUtils.nub compare (List.sort compare objs) in
    List.map (Hashtbl.find t.objects) ids

let drop_mine t id =
  match tank t id with
    | None -> ()
    | Some tank ->
	if tank#signal_mine then
	  create_object t
	    (new Mine.t (new_object_id t) (Object.Init { Mine.init_at = tank#location }))

let build t id kind distance =
  match tank t id, base t with
    | None, _ | _, None -> ()
    | Some tank, Some base ->
	let building_cost = 
	  match kind with
	    | Building.FakeBase _ -> 10
	    | Building.Wall _ -> 3 
	in
	let tank_face_vec =
	  let dir = -. tank#get_direction in
	    (cos dir, -.sin dir) 
	in
	let obj = 
	  new Building.t (new_object_id t)
	    (Object.Init
	       { (*Building.defaults with*)
		 Building.init_at = tank#location +|. tank_face_vec *|. Vector.v2 distance;
		 init_kind = kind;
		 init_hp = Building.kind_hp kind })
	in
	  if Building.has_space_for (obj :> Object.base)
	    (fun region -> 
	       List.filter
		 (fun o -> o#collidable) 
		 (GobaMap.objects_near t.map region @ objects_within t region)
	    )
	  then
	    if tank#signal_building building_cost then
	      create_object t obj

let build_fake_base t id =
  match base t with
    | None -> ()
    | Some base -> build t id (Building.FakeBase base#get_color) 50.0

let build_wall t id =
  match tank t id with
    | None -> ()
    | Some tank -> build t id (Building.Wall tank#get_direction) 15.0

let fire t id dir =
  match tank t id with
    | None -> ()
    | Some tank ->
	if tank#signal_firing then
	  (* shoot to the direction of the tank *)
	  let dir = 
	    let tank_face = 
	      let dir = -. tank#get_direction -. pi /. 2.0 in
		(cos dir, sin dir) in
	      dir **|. Vector.base tank_face in
	    create_object t 
	      (new Projectile.t (new_object_id t) 
		 (Object.Init 
		    { Projectile.init_at = tank#location +|. tank#turret_offset (atan2 (snd dir) (fst dir));
		      init_speed = (dir +|. tank#speed);
		      init_orientation = dir;
		      init_parent = tank#get_id } ));
	    fire_channel#play fire_sound

let stop_laser t =
  (match base t with Some base -> ignore (base#signal_laser false) | None -> ());
  match t.laser with
    | Some laser ->
	t.laser <- None;
	t.object_removed := laser :: !(t.object_removed);
	base_laser_channel#expire (Some 0.1);
    | None -> ()

let set_angle t angle =
  match base t with
    | None -> ()
    | Some base ->
	base#set_angle angle
	
let set_distance t distance =
  match base t with
    | None -> ()
    | Some base ->
	base#set_distance distance
	
let launch_base_bomb t =
  match base t with
    | None -> ()
    | Some base ->
	if base#signal_bomb then
	  let angle = base#get_angle /. 180.0 *. GobaMath.pi in
	  let distance = base#get_distance in
	  let ofs = Vector.v2 distance *|. (cos angle, -.sin angle) in
	    create_object t 
	      (new Bomb.t (new_object_id t)
		 (Object.Init
		    { Bomb.init_at = base#location +|. ofs;
		      init_time_left = Bomb.explosion_time }))

let fire_base_laser t enabled =
  match base t with
    | None -> ()
    | Some base ->
	match enabled with
	  | false -> stop_laser t
	  | true ->
	      let angle = base#get_angle /. 180.0 *. GobaMath.pi in
		if base#signal_laser true then
		  if t.laser = None then
		    let at = base#location in
		    let obj = new Laser.t (new_object_id t) 
		      (Object.Init 
			 { Laser.init_at = at;
			   init_angle = angle;
			   init_parent = base#get_id;
			   init_length = 0; })
		    in
		      create_object t obj;
		      base_laser_channel#play ~loops:(-1) base_laser_sound;
		      let oid = Object.to_object_id obj#get_id in
			t.laser <- Some oid
		  else
		    ()
		else
		  stop_laser t
		    
let add_object map region o =
  GobaUtils.iter2d
    (fun (x, y) ->
       Hashtbl.replace map (x, y)
	 (o::
	    try Hashtbl.find map (x, y)
	    with Not_found -> [])
    )
    region

let replace_object ?(remove_only=false) t (o:Object.base) =
  let id = Object.to_object_id o#get_id in
  let region = region t (fst o#collision_area) (0, 0) in
  let w, h = t.map.GobaMap.width, t.map.GobaMap.height in
  let old_tiles = 
    try 
      let ts = Hashtbl.find t.object_old_map id in
	ts
    with Not_found -> [] 
  in
    List.iter 
      (fun (x, y) ->
	 t.object_map.(x + y * w) <-
	   List.filter ((<>) id) t.object_map.(x + y * w)
      )
      old_tiles;
    if remove_only then
      Hashtbl.remove t.object_old_map id
    else
      let ats = 
	GobaUtils.foldl2d 
	  (fun (x, y) ats ->
	     t.object_map.(x + y * w) <- id::t.object_map.(x + y * w);
	     (x, y)::ats
	  )
	  []
	  region 
      in
	Hashtbl.replace t.object_old_map id ats

let tick t delta =
  let objs ?except collidables = objects ?except ~collidables t in
  let () = 
    List.iter
      (replace_object t)
      (objects ~changed_since:t.object_map_serial t);
    t.object_map_serial <- Object.next_serial ()
  in
  let world collidables self = 
    let normal_objs = lazy ((objs ~except:self collidables) :> Object.t list) in
      (object
	 method explode =
	   (*	  (fun at -> ());*)
	  (fun ?(sound=true) ?(kind=ExplosionKind.Particle) at burst ->
	     let id = new_object_id t in
	     let obj = (new Explosion.t id (Object.Init { Explosion.init_at = at;
							  init_burst = burst;
							  init_sound = sound;
							  init_damage = Explosion.default_damage;
							  init_kind = kind }) :> Object.base) in
	       obj#set_self_owned true;
	       Client.create_object t.client 
		 (new_object_info t obj);
	       Hashtbl.add t.objects id obj
	  )
	 method collision = 
	   (fun obj filter -> 
	      Collision.collisions2 
		(obj : #Collision.t) 
		(List.rev_append 
		   (Lazy.force normal_objs) 
		   (GobaMap.objects_near t.map (fst obj#collision_area))
		)
		filter
	   );
	 method env =
	   (fun obj filter ->
	      Collision.collisions 
		(obj : #Collision.t) 
		(GobaMap.objects_near t.map (fst obj#collision_area)) 
		filter);
	 method objs =
	   (fun ((x1, y1), (x2, y2)) filter ->
(*	      Printf.printf "Objects within %f,%f-%f,%f\n%!" x1 y2 x2 y2; *)
	      let filter o = 
		let (x, y) = o#location in
(*		  Printf.printf "Checking %s (%s) %b %b %b\n%!" (Object.string_of_id o#get_id)
		    (Vector.f_to_string o#location)
		    (within x (x1, x2))
		    (within y (y1, y2))
		    (filter o);*)
	          within x (x1, x2) && within y (y1, y2) && filter o
	      in
		List.filter filter 
		  (objects_within t ((x1, y1), (x2, y2)))
		@ List.filter filter (GobaMap.objects_near t.map ((x1, y1), (x2, y2)))
	   );
	 method report_damage =
	   (fun obj_damage -> 
	      Client.damage_object t.client obj_damage
	   );
	 method channel = 
	   (fun ch -> new SfxSdl.channel ch)

	 method find_object id =
	   try Some (Hashtbl.find t.objects (Object.to_object_id id))
	   with Not_found -> None

	 method create_object o =
	   create_object t (o (new_object_id t))
       end)
  in
    (match t.laser with
       | None -> ()
       | Some id ->
	   if not (Hashtbl.mem t.objects id) || not (Hashtbl.find t.objects id)#is_alive then
	     begin
	       t.laser <- None;
	       base_laser_channel#expire (Some 0.1);
	     end
    );

    List.iter
      (fun o ->
	 o#tick delta (world true o)
      ) (objs true);

    List.iter
      (fun o ->
	 o#tick delta (world true o)
      ) (objs false)

      (*  List.iter (fun o -> o#tick delta) (objects t)*)

let center_view at size =
  at -|. size /|. Vector.dup 2.0, size

let handle_io t =
  let now = Unix.gettimeofday () in
    Client.handle_io t.client;
    if now > t.last_sync +. sync_interval then
      begin
	begin
	  match t.pending_ping with
	    | false ->
		t.ping_value := None;
		t.pending_ping <- true;
		ignore (Client.ping t.client t.ping_value)
	    | true ->
		match !(t.ping_value) with
		  | None -> ()
		  | Some latency -> 
		      (*Printf.printf "Latency: %f\n%!" latency;*)
		      t.latency := latency *. 0.1 +. !(t.latency) *. 0.9;
		      Client.report_latency t.client latency;
		      t.pending_ping <- false
	end;
	t.last_sync <- now;
	(* Printf.printf "%d objects\n%!" (List.length (objects ~self_owned:true ~changed_since:t.last_sync_serial t));*)
	let remove_ids =
	  List.fold_left
	    (fun del o ->
	       if not o#is_alive then
		 o#get_id::del
	       else
		 del) 
	    []
	    (objects ~self_owned:true ~changed_since:t.last_sync_serial t) 
	in
	  List.iter 
	    (fun oid ->
	       try
		 let obj = Hashtbl.find t.objects oid in
		   if obj#self_owned then
		     Client.destruct_object t.client (Object.ObjectId oid);
		   replace_object ~remove_only:true t obj;
		   Hashtbl.remove t.tanks oid;
		   Hashtbl.remove t.bases oid;
		   Hashtbl.remove t.objects oid;
		   obj#finish;
	       with Not_found -> ()
		 (** FIXME: why does Not_found get thrown when we hit a mine? *)
	    )
	    (* objects may have died and thus they may be in the remove_ids-list, thus we want to filter duplicates *)
	    (GobaUtils.nub compare (List.sort compare (List.map Object.to_object_id remove_ids) @ !(t.object_removed)));
	  t.object_removed := [];
	  List.iter 
	    (fun o ->
	       if o#is_alive then
		 Client.report_object t.client (o#get_id, o#get_state 0.0)
	    )
	    (objects ~self_owned:true ~changed_since:t.last_sync_serial t);
	  t.last_sync_serial <- Object.next_serial ()
      end
    (*Printf.printf "%d objects\n%!" (List.length (objects t))*)

let section_of_coords t at =
  let size = 
    (Vector.float2 (t.map.GobaMap.width, t.map.GobaMap.height) +|.
	 (1.0, 1.0)) *|.
	Vector.float2 (t.map.GobaMap.tile_size)
  in
  let section = Vector.int2 (at /|. size *|. (10.0, 10.0)) in
    Printf.sprintf "%c%c" (Char.chr (Char.code 'A' + fst section)) (Char.chr (Char.code '0' + snd section))

let chat t msg =
  Client.chat t.client msg

let in_game t =
  match base t with
    | Some base -> true
    | None -> false

let adjust_endgame_view t adjust =
  t.endgame_view <- adjust t.endgame_view

let render_chat t (gfx:Gfx.t) =
  ignore (
    List.fold_left 
      (fun at (_, surface) ->
	 gfx#blit at surface;
	 at -| (0, 20)
      )
      (10, 550)
      !(t.messages)
  )

let shrink_region (a, b) by =
  (a +|. by, b -|. by)

let draw_guide 
    (gfx:Gfx.t)
    at
    (disp_x, disp_y)
    (disp_w, disp_h)
    =
  let color1 = (240, 220, 150) in
  let color2 = (220, 150, 240) in
  let clip = ((disp_x, disp_y), (disp_x + disp_w, disp_y + disp_h)) in
    List.iter
      (fun ang_deg ->
	 let ang = -. ang_deg /. 180.0 *. GobaMath.pi in
	 let min_step = 0.003 in
	 let ang0 = ang -. min_step in
	 let ang1 = ang +. min_step in
	 let ang2 = ang +. GobaMath.pi -. min_step in
	 let ang3 = ang +. GobaMath.pi +. min_step in
	 let ofs ?(dist=500.0) ang =
	   at +| Vector.int2 (cos ang *. dist, sin ang *. dist);
	 in
	 let label ang ang_deg =
	   let surface = gfx#text_surface ~color:color1 ~size:19 (Printf.sprintf "%.1f" ang_deg) in
	   let (w, h) = Gfx.surface_dims surface in 
	     gfx#blit (ofs ~dist:120.0 (ang) -| (w, h) /| (2, 2)) surface
	 in
	   polygon ~clip gfx color1
	     (List.rev [ofs ang0; ofs ang1; ofs ang2; ofs ang3]);
	   label ang ang_deg;
	   label (ang +. GobaMath.pi) (ang_deg +. 180.0);
      )
      [0.0; 90.0; 45.0; 90.0 +. 45.0;
       45.0 /. 2.0; 45.0 +. 45.0 /. 2.0; 90.0 +. 45.0 /. 2.0; 90.0 +. 45.0 +. 45.0 /. 2.0];
    gfx#with_clip_rect
      (Sdlvideo.rect disp_x disp_y disp_w disp_h)
      (fun () -> 
	 List.iter 
	   (fun radius ->
	      circle gfx color2 1 20 at radius;
	      gfx#write ~color:color2 ~size:19 (at +| (5, int_of_float radius)) 
		(Printf.sprintf "%.0f" radius)
	   )
	   [50.0; 100.0; 150.0; 200.0; ]
      )


let disp_base = (10, 10)
let disp_width, disp_height = (640, 400)
let tank_info_bar_width = 50
let tank_display_size = (180.0, 195.0)
let main_display_size = (300.0, 
			 float disp_height)
let base_info_bar_at = (int_of_float (10.0 +. fst tank_display_size +. float tank_info_bar_width), 
			int_of_float (snd main_display_size))
let main_display_at = disp_base +| (int_of_float (fst tank_display_size +. float tank_info_bar_width) + 10, 0)

let tanks =
  [TankId 0, 
   (tank_info_bar_width, 0), (fun (x, y) -> (x - tank_info_bar_width, y));

   TankId 1, 
   (20 + int_of_float (fst tank_display_size +. fst main_display_size +. float tank_info_bar_width), 0),
   (fun (x, y) -> (x + int_of_float (fst tank_display_size), y));

   TankId 2, 
   (tank_info_bar_width, disp_height - int_of_float (snd tank_display_size)),
   (fun (x, y) -> (x - tank_info_bar_width, y));

   TankId 3, 
   (20 + int_of_float (fst tank_display_size +. fst main_display_size +. float tank_info_bar_width), disp_height - int_of_float (snd tank_display_size)),
   (fun (x, y) -> (x + int_of_float (fst tank_display_size), y));
  ]

let regions =
  [(main_display_at, Vector.int2 main_display_size, 
    (fun t at -> 
       match base t with
	 | None -> None
	 | Some base ->
	     Some (Vector.float2 at +|. fst (center_view base#location main_display_size))))
  ] @ 
    List.map
    (fun (id, base, _) ->
       (base +| disp_base, Vector.int2 tank_display_size,
	fun t at ->
	  match tank t id with
	    | None -> None
	    | Some tank ->
		Some (Vector.float2 at +|. fst (center_view tank#location tank_display_size))
       )
    )
    tanks

let set_tank_goal t id at =
  match tank t id with
    | None -> ()
    | Some tank -> tank#set_goal at

let screen_to_map_coord t ((sx, sy) as screen_at) =
  let region = 
    try
      Some (List.find 
	      (fun ((x, y), (w, h), _) ->
		 within sx (x, x + w) && within sy (y, y + h)
	      ) regions)
    with Not_found -> None
  in
    match region with
      | None -> None
      | Some (base, _, f) ->
	  f t (screen_at -| base)

let pointer_object pointer =
  match pointer with
    | None -> None
    | Some at ->
	Some
	  (object
	     method location = at
	     method render gfx (_:Object.render_flags) at = 
	       [(0, fun () ->
		   Bitmaps.blit gfx at @@ Bitmaps.bitmap @@ B.native @@ Bitmaps.move (-60, -60) @@ Bitmaps.auto_crop @@ Bitmaps.file "crosshair-0001")]
	   end
	  ) 

let render_base t (gfx:Gfx.t) want_guide (base:Base.t) objects pointer =
  (* main view *)
  let base_at = base#location in
    if try Sys.getenv "GOBA_DEBUG" = "0" with Not_found -> true then
      begin
	let view = center_view base_at main_display_size in
	let labels = 
	  let label_offset = (-5, -35) in
	  let (bx1, by1) = fst view +|. (0.0, 0.0) -|. Vector.float2 label_offset in
	  let (bx2, by2) = fst view +|. snd view -|. (13.0, 20.0) -|. Vector.float2 label_offset  in
	    List.concat (
	      List.map
		(fun (id, label) ->
		   match tank t id with
		     | Some t ->
			 let (x, y) = t#location in
			 let at = (GobaMath.clamp bx1 bx2 x, GobaMath.clamp by1 by2 y) in
			   [object
			      method location = at +|. Vector.float2 label_offset;
			      method render (gfx:Gfx.t) flags at = [(0, fun () -> gfx#write ~size:20 ~color:(128, 255, 128) at label)]
			    end]
		     | None -> [])
	      [TankId 0, "1"; TankId 1, "2"; TankId 2, "3"; TankId 3, "4"]
	    )
	in
	  GobaMap.render_map t.map gfx [`Labels] main_display_at 
	    view objects (labels @ (match pointer with None -> [] | Some p -> [p]));
	  fade_rect gfx (main_display_at, Vector.int2 (snd view));
	  if want_guide then
	    draw_guide gfx 
	      (Vector.int2 (base#location -|. fst view) +| main_display_at)
	      main_display_at
	      (Vector.int2 (snd view))
	  else
	    ();
      end;
    GobaUtils.iter1d 
      (fun n ->
	 Bitmaps.blit gfx (base_info_bar_at +| disp_base +| Vector.v2 n *| (10, 0) +| (20, 20))
	   (Lazy.force bomb_symbol)
      )
      (0, base#get_bombs);
    gfx#rectangle 
      (0xcf, 0x90, 0x03)
      (base_info_bar_at +| disp_base +| (2, 20)) 
      (int_of_float (base#get_laser_load *. 10.0), 5)
    ;
    gfx#write ~size:20 (base_info_bar_at +| disp_base) 
      (Printf.sprintf 
	 "%d %s" 
	 (int_of_float base#get_hp)
	 (section_of_coords t base#location)
      );
    Bitmaps.blit gfx (base_info_bar_at +| disp_base +| (130, 25)) (Lazy.force angle_symbol);
    gfx#write ~size:20 (base_info_bar_at +| disp_base +| (145, 10)) (Printf.sprintf "%.1f%c" base#get_angle degree);
    Bitmaps.blit gfx (base_info_bar_at +| disp_base +| (260, 10)) (Lazy.force distance_symbol);
    gfx#write ~size:20 (base_info_bar_at +| disp_base +| (230, 10)) (Printf.sprintf "%.0fpx" base#get_distance)

let render_lights t (gfx:Gfx.t) =
  ignore (
    List.fold_left
      (fun at (id, color, base) ->
	 let brightness =
	   let alive = 
	     try (Hashtbl.find t.objects base)#is_alive
	     with Not_found -> false
	   in
	     if alive then
	       1.0
	     else
	       0.4
	 in
	 let light = 
	   Bitmaps.bitmap 
	   @@ B.native
	   @@ Bitmaps.brightness brightness
	   @@ Bitmaps.colored color 
	   @@ Bitmaps.move (-60, -60)
	   @@ Bitmaps.auto_crop 
	   @@ Bitmaps.no_cache
	   @@ Bitmaps.file "signal_light"
	 in
	   Bitmaps.blit gfx (base_info_bar_at +| (350 + at, 25)) light;
	   at + 30
      )
      0
      t.players;
  )

let render_tank t (gfx:Gfx.t) objects cur_tanks pointer (id, at, info_at_f) =
  let info_at = info_at_f at +| disp_base in
    match tank t id with
      | None -> ()
      | Some tank ->
	  (*gfx#write ~size:20 (600, text_y) (Vector.f_to_string tank#location);*)
	  GobaMap.render_map t.map gfx [`LabelsExcept [tank#get_id]] (disp_base +| at) 
	    (center_view tank#location tank_display_size) objects 
	    (match pointer with None -> [] | Some p -> [p]);
	  let _ =
	    let damage = tank#get_damage_counter in
	      if damage > 0.0 then
		let view_alpha, view_bg = 
		  int_of_float (255.0 *. 
				  min 
				  1.0 
				  damage),
		  (255, 0, 0)
		in
		  gfx#blit (disp_base +| at) (Help.alpha_rectangle gfx#get_surface view_alpha view_bg (Vector.int2 (tank_display_size)));
	  in
	  let bg_color = 
	    if List.exists ((=) id) cur_tanks then
	      (0xad, 0x8a, 0xd6)
	    else
	      (0x7c, 0x5c, 0xa0)
	  in
	  let color = if tank#is_alive then (128, 255, 128) else (64, 128, 64) in
	  let info ?(size=20) at format =
	    Printf.kprintf (fun msg -> gfx#write ~color ~size (info_at +| at) msg) format
	  in
	  let hp_ratio = GobaMath.f_scale 0.0 Tank.max_hp tank#get_hp in
	  let hp_pixels = int_of_float (snd tank_display_size *.  hp_ratio) in
	  let display_height = int_of_float (snd tank_display_size) in
	  let hp_color = ipolate3 hp_ratio (200, 0, 0) (0, 200, 0) in
	    gfx#rectangle bg_color info_at (tank_info_bar_width, display_height);
	    gfx#rectangle hp_color (info_at +| (0, display_height - hp_pixels)) (tank_info_bar_width, hp_pixels);
	    (*gfx#blit info_at (alpha_rectangle gfx#get_surface 128 (255, 255, 255) (30, 30));*)
	    info ~size:40 (0, 0) "#%d" (match id with TankId i -> i + 1);
	    (*info (0, 40) "%d" tank#get_hp;*)
	    info (0, 60) "%s" (section_of_coords t tank#location);
	    info (0, 80) "%d" tank#get_ammo;
	    info (0, 100) "%d" tank#get_mines;
	    info (0, 120) "%d" tank#get_building_points;

	    render_chat t gfx;
	    fade_rect gfx
	      ((min (fst info_at) (fst (disp_base +| at)), snd info_at),
	       Vector.int2 tank_display_size +| (tank_info_bar_width, 0))

let render t (gfx:Gfx.t) want_guide cur_tanks =
  gfx#blit (0, 0) (Lazy.force background);
  match base t with 
    | Some base ->
	let objects = objects t in
	let pointer =
	  let (m_x, m_y, _) = Sdlmouse.get_state () in
	    pointer_object (screen_to_map_coord t (m_x, m_y))
	in
	  render_base t gfx want_guide base objects pointer;
	  render_lights t gfx;
	  List.iter (render_tank t gfx objects cur_tanks pointer) tanks
    | None -> 
	GobaMap.render_map t.map gfx [] disp_base
	  (center_view t.endgame_view (780.0, 580.0)) (objects t) [];
	gfx#write ~size:20 (20, 20)
	  (Printf.sprintf 
	     "%s" 
	     (section_of_coords t t.endgame_view)
	  );
	render_chat t gfx;



	 
