open VecOps
open GobaMath

let gfx = new Gfx.t ~size:(800, 600) ()

type setup = { setup_meta : bool;
	       setup_host : string option }

let timer_hz = 120.0
let refresh_hz = try float_of_string (Sys.getenv "HZ") with _ -> 30.0
let world_tick_hz = 60.0

let timer_interval = 1.0 /. timer_hz
let refresh_interval = 1.0 /. refresh_hz
let world_tick_interval = 1.0 /. world_tick_hz

let background =
  let background_picture = lazy (Bitmaps.bitmap (Bitmaps.file "begin")) in
  let logo = 
    let logo_width = 513 in
      lazy (Bitmaps.bitmap (Bitmaps.move (-logo_width / 2, 0) (Bitmaps.realpha 0.7 (Bitmaps.file "goba-logo")))) in
    fun (gfx : Gfx.t) ->
      Bitmaps.blit gfx (0, 0) (Lazy.force background_picture);
      Bitmaps.blit gfx (400, 20) (Lazy.force logo)

let () = 
  Sdlmixer.open_audio ();
  ignore (Sdlmixer.allocate_channels 16);
  if false then
    let music = Sdlmixer.load_music "goba.ogg" in
      Sdlmixer.play_music ~loops:0 music


let pending_mutex = Mutex.create ()
let pending_timers = ref 0

let rec ticker () =
  let extra = float !pending_timers *. 0.01 in
    (try ignore (Unix.select [] [] [] (1.0 /. refresh_hz +. extra)) with exn -> ());
    Sdlevent.add [Sdlevent.USER 0];
    Mutex.lock pending_mutex; incr pending_timers; Mutex.unlock pending_mutex;
    ticker ()

let tank_keys = [Sdlkey.KEY_1, Game.TankId 0;
		 Sdlkey.KEY_2, Game.TankId 1;
		 Sdlkey.KEY_3, Game.TankId 2;
		 Sdlkey.KEY_4, Game.TankId 3]

let endgame_keys = [Sdlkey.KEY_DOWN, (fun delta at -> at +|. Vector.v2 delta *|. (0.0, 300.0));
		    Sdlkey.KEY_UP, (fun delta at -> at +|. Vector.v2 delta *|. (0.0, -.300.0));
		    Sdlkey.KEY_LEFT, (fun delta at -> at +|. Vector.v2 delta *|. (-.300.0, 0.0));
		    Sdlkey.KEY_RIGHT, (fun delta at -> at +|. Vector.v2 delta *|. (300.0, 0.0));
		   ]

let firedirs = [Sdlkey.KEY_q, 360 - 90 - 45;
		Sdlkey.KEY_w, 360 - 90;
		Sdlkey.KEY_e, 360 - 45;

		Sdlkey.KEY_a, 180;
		Sdlkey.KEY_s, 90;
		Sdlkey.KEY_d, 0;

		Sdlkey.KEY_z, 90 + 45;
		Sdlkey.KEY_x, 90;
		Sdlkey.KEY_c, 45]

let dirs = [Sdlkey.KEY_DOWN, (fun delta game tank -> Game.accel_tank game tank (-. (150.0 *. delta)));
	    Sdlkey.KEY_UP, (fun delta game tank -> Game.accel_tank game tank (150.0 *. delta));
	    Sdlkey.KEY_LEFT, (fun delta game tank -> Game.turn_tank game tank (90.0 *. -.pi /. 180.0 *. delta));
	    Sdlkey.KEY_RIGHT, (fun delta game tank -> Game.turn_tank game tank (90.0 *. pi /. 180.0 *. delta));
	   ]

let avg_dir d1 d2 =
  let d1 = mod_float d1 pi2 in
  let d2 = mod_float d2 pi2 in
  let d2 = if d2 < d1 then d2 +. pi2 else d2 in
  let d1 = 
    if d1 < d2 -. pi then 
      d1 +. pi2 
    else
      d1
  in
    (d1 +. d2) /. 2.0

let run_generic
    ?(user_handler=fun continue _ -> continue ()) 
    ?(event_handler=fun continue _ -> continue ())
    ~render
    ~exit
    ()
    =
  let rec loop prev_time = 
    match Sdlevent.wait_event () with
      | Sdlevent.QUIT -> exit
      | Sdlevent.USER 0 ->
	  let () = 
	    Mutex.lock pending_mutex; 
	    let _v = !pending_timers in
	      decr pending_timers; 
	      Mutex.unlock pending_mutex
	  in
	    user_handler
	      (fun () -> 
		 let now = Unix.gettimeofday () in
		 let delta = now -. prev_time in
		   render gfx delta;
		   gfx#sync ();
		   loop now)
	      0
      | Sdlevent.USER n ->
	  user_handler (fun () -> loop prev_time) n
      | event ->
	  event_handler
	    (fun () -> loop prev_time)
	    event
  in
    loop (Unix.gettimeofday ())

let prompt message =
  let input = ReadInput.create (10, 200) in
    run_generic
      ~event_handler:(
	fun continue event ->
	  match ReadInput.handle input event with
	    | Some message -> Some message
	    | None -> continue ())
      ~exit:None
      ~render:(
	fun gfx delta ->
	  gfx#clear ();
	  gfx#write (10, 50) message;
	  ReadInput.render input gfx;
      )
      ()

let run_menu 
    ?(user_handler=fun continue _ -> continue ()) 
    ?(background=(fun gfx -> gfx#clear ()))
    ?(foreground=(fun _ -> ()))
    menu callback =
  run_generic
    ~user_handler
    ~event_handler:(
      fun continue event ->
	match event with 
	  | Sdlevent.KEYDOWN { Sdlevent.keysym = Sdlkey.KEY_DOWN } ->
	      Menu.choose_next menu;
	      continue ()
	  | Sdlevent.KEYDOWN { Sdlevent.keysym = Sdlkey.KEY_UP } ->
	      Menu.choose_prev menu;
	      continue ()
	  | Sdlevent.KEYDOWN 
	      { Sdlevent.keysym = (Sdlkey.KEY_SPACE | Sdlkey.KEY_RETURN) } ->
	      callback (fun () -> continue ()) (Menu.choose menu)
	  | _ -> continue ()
    )
    ~render:(
      fun gfx delta ->
	Menu.tick menu delta;
	background gfx;
	Menu.render menu gfx;
	foreground gfx;
    )
    ~exit:()
    ()
      
let rec player_menu ?(selected=0) client game_loop =
  let unhook () = 
    Client.set_available_players_handler client Client.nothing
  in
    Client.set_available_players_handler client
      (fun available ->
	 Sdlevent.add [Sdlevent.USER 1]
      );
    let available = Client.available client in
    let player n =
      Menu.entry (Printf.sprintf "Player %d" (n + 1)) n
    in
    let menu =
      (Menu.create ~selected
	 (List.map player (
	    List.filter 
	      (fun id -> List.mem (Player.Id id) available)
	      [0; 1; 2; 3]
	  ))
	 ~justify:(Menu.align_south ~marginal:50)
      )
    in
      run_menu 
	~background
	~user_handler:(fun continue n ->
			 if n = 1 then
			   player_menu ~selected:(Menu.currently_selected menu) client game_loop
			 else
			   begin
			     Client.handle_io client;
			     continue ()
			   end
		      )
	menu
	(fun _ player ->
	   unhook ();
	   game_loop client player
	)

let rec connect_menu setup ?(message="") ?(selected=0) game_loop =
  let last_query = Unix.gettimeofday () in
  let servers = 
    if setup.setup_meta then
      Meta.list_servers (Meta.create "http://modeemi.cs.tut.fi/~flux/update.cgi") 
    else
      [] 
  in
  let menu = 
    (Menu.create ~selected ~font:(Gfx.get_font ~size:30 ())
       ~justify:(Menu.align_south ~marginal:50)
       ([Menu.entry "Connect localhost" (Some ("localhost", 61753));
	 Menu.entry "Connect.." None]
	@ 
	 List.map 
	 (fun { Meta.server_name = name; server_addr = (host, port) } -> 
	    let hostname = 
	      try
		(Unix.gethostbyaddr host).Unix.h_name
	      with Not_found -> Unix.string_of_inet_addr host
	    in
	      Menu.entry ("Connect " ^ name ^ "@" ^ hostname) (Some (Unix.string_of_inet_addr host, port))
	 ) 
	 servers
       ))
  in
    run_menu
      ~background
      ~foreground:(fun gfx ->
		     if message <> "" then
		       gfx#write (20, 20) message
		  )
      ~user_handler:(fun continue n ->
		       let now = Unix.gettimeofday () in
			 if n = 0 && now -. last_query > 10.0 then
			   connect_menu setup ~selected:(Menu.currently_selected menu) game_loop
			 else
			   continue ())
      menu
      (fun _ address ->
	 let connect (server, port) failure =
	   let client = 
	     try
	       let host = Unix.gethostbyname server in
	       let server = (Unix.ADDR_INET (host.Unix.h_addr_list.(0), port)) in
		 Some (Client.create server)
	     with 
	       | Unix.Unix_error _ -> None
	       | Not_found -> None
	   in
	     match client with
	       | Some client -> player_menu client game_loop
	       | None -> failure ()
	 in
	 let fail ?(message="Failed to connect") () = connect_menu setup ~message ~selected:(Menu.currently_selected menu) game_loop in
	   match address with
	     | Some address -> 
		 connect address fail
	     | None -> 
		 match prompt "Connect to address:" with
		   | None -> ()
		   | Some address ->
		       let address, port =
			 try 
			   let idx = String.rindex address ':' in
			     String.sub address 0 idx, int_of_string (String.sub address (idx + 1) (String.length address - idx - 1))
			 with 
			   | Not_found -> address, 61753
			   | Failure "int_of_string" -> address, 61753 (*fail ~message:"Invalid port" ()*)
		       in
			 connect (address, port) fail
      )


let main_menu setup game_loop =
  run_menu 
    ~background
    (Menu.create [Menu.entry "Join game" `Start;
		  Menu.entry "Start server" `None;
		  Menu.entry "Quit" `Quit]
(*       ~justify:(fun ~menu_size ~screen_size ~size ~at -> 
		   Menu.center ~menu_size ~screen_size ~size ~at +| (0, 250))*)
       ~justify:(Menu.align_south ~marginal:50)
       (*~justify:Menu.center*)
    )
    (fun continue chosen ->
       match chosen with
	 | `Start -> connect_menu setup game_loop
	 | `Quit -> ()
	 | `None -> continue ()
    )

type input = { input_label : Sdlvideo.surface;
	       input_read : ReadInput.t;
	       input_handler : (game_state -> string -> game_state) }

and game_state = { cur_tanks		: Game.tank_id list;
		   world_time		: float;
		   world_changed	: bool;
		   refresh_time		: float;
		   input		: input option; 
		   guide		: bool; }

let create_input label_size = ReadInput.create (0 + fst label_size, 420)

let create_chat (gfx:Gfx.t) game =
  let handler state input =
    Game.chat game input;
    state
  in
  let label = gfx#text_surface ~size:25 ">" in
    { input_label = label;
      input_read = create_input (Gfx.surface_dims label);
      input_handler = handler }

let create_angle (gfx:Gfx.t) game =
  let handler state input =
    ( match
	try Some (float_of_string input)
	with Failure _ -> None
      with
	| Some ang -> Game.set_angle game ang
	| None -> () );
      { state with guide = false }
  in
  let label = gfx#text_surface ~size:25 "angle:" in
    { input_label = label;
      input_read = create_input (Gfx.surface_dims label);
      input_handler = handler }

let create_distance (gfx:Gfx.t) game =
  let handler state input =
    ( match
	try Some (float_of_string input )
	with Failure _ -> None
      with
	| Some d -> Game.set_distance game d
	| None -> () );
    { state with guide = false }
  in
  let label = gfx#text_surface ~size:25 "distance:" in
    { input_label = label;
      input_read = create_input (Gfx.surface_dims label);
      input_handler = handler }

let create_command (gfx:Gfx.t) game =
  let handler state input =
    let command, args =
      match Pcre.split ~pat:" " input with
	| [] -> ("", [])
	| command::args -> (command, args)
    in
      List.iter
	(fun id -> match Game.tank game id with None -> () | Some tank -> tank#command command args)
	state.cur_tanks;
      state
  in
  let label = gfx#text_surface ~size:25 "command:" in
    { input_label = label;
      input_read = create_input (Gfx.surface_dims label);
      input_handler = handler }

let game_loop server player_id =
  let frame = ref 0 in
  let game = Game.create server player_id in
  let rec loop prev_time state =
    match Sdlevent.wait_event () with
      | Sdlevent.QUIT -> 
	  Printf.printf "Quit\n%!";
      | Sdlevent.KEYDOWN ({ Sdlevent.keysym = Sdlkey.KEY_RETURN } as kev) when kev.Sdlevent.keymod land (Sdlkey.kmod_lalt lor Sdlkey.kmod_ralt lor Sdlkey.kmod_lmeta lor Sdlkey.kmod_rmeta) <> 0 ->
	  ignore (Sdlwm.toggle_fullscreen ());
	  loop prev_time state
      | Sdlevent.KEYDOWN { Sdlevent.keysym = Sdlkey.KEY_RETURN } when state.input = None ->
	  loop prev_time { state with input = Some (create_chat gfx game) }
      | Sdlevent.KEYDOWN { Sdlevent.keysym = Sdlkey.KEY_p } when state.input = None ->
	  loop prev_time { state with input = Some (create_angle gfx game); guide = true; }
      | Sdlevent.KEYDOWN { Sdlevent.keysym = Sdlkey.KEY_o } when state.input = None ->
	  loop prev_time { state with input = Some (create_distance gfx game); guide = true; }
      | Sdlevent.KEYDOWN ({ Sdlevent.keysym = Sdlkey.KEY_c } as kev) when state.input = None && kev.Sdlevent.keymod land (Sdlkey.kmod_ctrl) <> 0 ->
	  loop prev_time { state with input = Some (create_command gfx game) }
      | Sdlevent.MOUSEBUTTONDOWN { Sdlevent.mbe_x = x; mbe_y = y } when state.cur_tanks <> [] ->
	  ( match Game.screen_to_map_coord game (x, y) with
	      | None -> ()
	      | Some at -> 
		  match Game.tank game (List.hd state.cur_tanks) with
		    | None -> ()
		    | Some tank ->
			let base = tank#get_goal in
			  List.iter 
			    (fun id -> 
			       match Game.tank game id with
				 | None -> ()
				 | Some tank -> tank#set_goal (at +|. tank#get_goal -|. base))
			    state.cur_tanks );
	  loop prev_time state
      | (Sdlevent.KEYDOWN _) as event when state.input <> None ->
	  begin
	    match state.input with
	      | None -> assert false
	      | Some { input_read = input; input_handler = handler } ->
		  match ReadInput.handle input event with
		    | Some msg -> 
			let state = handler state msg in
			  loop prev_time { state with input = None }
		    | None ->
			loop prev_time state
	  end;
      | Sdlevent.KEYDOWN { Sdlevent.keysym = key; keymod = mods } when List.exists (fun (key', _) -> key = key') tank_keys ->
	  let state = 
	    { state with cur_tanks =
		List.fold_left 
		  (fun cur_tanks (key', tank) ->
		     if key' = key then
		       if mods land Sdlkey.kmod_shift <> 0 then
			 if List.exists ((=) tank) cur_tanks then
			   List.filter ((<>) tank) cur_tanks
			 else
			   tank::cur_tanks
		       else
			 [tank]
		     else
		       cur_tanks
		  )
		  state.cur_tanks
		  tank_keys
	    }
	  in
	    loop prev_time state
      | Sdlevent.KEYDOWN { Sdlevent.keysym = Sdlkey.KEY_b } ->
	  Game.launch_base_bomb game;
	  loop prev_time state
      | Sdlevent.KEYDOWN { Sdlevent.keysym = Sdlkey.KEY_m } ->
	  List.iter (Game.drop_mine game) state.cur_tanks;
	  loop prev_time state
      | Sdlevent.KEYDOWN { Sdlevent.keysym = Sdlkey.KEY_g } ->
	  List.iter (Game.build_fake_base game) state.cur_tanks;
	  loop prev_time state
      | Sdlevent.KEYDOWN { Sdlevent.keysym = Sdlkey.KEY_t } ->
	  List.iter (Game.build_wall game) state.cur_tanks;
	  loop prev_time state
      | Sdlevent.USER 0 ->
	  let () = 
	    Mutex.lock pending_mutex; 
	    let _v = !pending_timers in
	      decr pending_timers; 
	      Mutex.unlock pending_mutex
	  in
	  let handle_controls delta =
	    let projectile_speed = 300.0 in
	      if not (List.fold_left 
			(fun was (key, f) ->
			   if Sdlkey.is_key_pressed key then
			     begin
			       List.iter
				 (fun tank ->
(*				    Game.accel_tank game tank (vec *|. Vector.dup (accel *. delta))*)
				    f delta game tank;
				 )
				 state.cur_tanks;
			       true
			     end
			   else
			     was)
			false
			dirs) then
		List.iter 
		  (fun tank ->
		     (match Game.tank game tank with None -> () | Some tank -> tank#decelerate (1.0 ** delta)))
		  state.cur_tanks;
	      let fire_dir = 
		let dirs = 
		  List.map (fun (_, dir) -> float dir /. 180.0 *. 4.0 *. atan 1.0) (
		    List.filter (fun (key, dir) -> Sdlkey.is_key_pressed key) firedirs
		  )
		in
		  if dirs <> [] then
		    let min_dir = List.fold_left min (List.hd dirs) (List.tl dirs) in
		    let max_dir = List.fold_left max (List.hd dirs) (List.tl dirs) in
		    let dir = avg_dir min_dir max_dir +. (Random.float 1.0 -. 0.5) /. pi2 in
		      Some (cos dir, sin dir)
		  else
		    None
	      in
		(match fire_dir with
		   | None -> ()
		   | Some dir -> List.iter (fun tank -> Game.fire game tank (dir *|. Vector.dup projectile_speed)) state.cur_tanks);
		if Sdlkey.is_key_pressed Sdlkey.KEY_l then
		  Game.fire_base_laser game true
		else
		  Game.fire_base_laser game false
	  in
	    
	  let handle_endgame_controls delta =
	    let adjust =
	      List.fold_left
		(fun dir (key, adj) ->
		   if Sdlkey.is_key_pressed key then
		     adj delta dir
		   else
		     dir
		)
		(0.0, 0.0)
		endgame_keys
	    in
	      Game.adjust_endgame_view game ((+|.) adjust)
	  in

	  let step delta =
	    if state.input = None then
	      if Game.in_game game then
		handle_controls delta
	      else
		handle_endgame_controls delta;
	    Game.tick game delta;
	  in
	  let now = Unix.gettimeofday () in
	  let delta = now -. prev_time in
	  let rec ticktick (changed, left) =
	    if left >= world_tick_interval then
	      begin
		step world_tick_interval;
		ticktick (true, left -. world_tick_interval)
	      end
	    else
	      changed, left
	  in
	    Game.handle_io game;
	    let world_changed, world_time = ticktick (state.world_changed, state.world_time +. delta) in
	    let world_changed, refresh_time =
	      let t = state.refresh_time +. delta in
		if t >= refresh_interval && world_changed then
		  begin
		    (*Printf.printf "frame %d\n%!" !frame;*)
		    incr frame;
		    (*gfx#clear ();*)
		    Game.render game gfx state.guide (state.cur_tanks);
		    (match state.input with 
			 Some { input_label = surface; input_read = input } -> 
			   gfx#blit (0, 426) surface;
			   ReadInput.render input gfx 
		       | None -> ());
		    if Sdlkey.is_key_pressed Sdlkey.KEY_F1 then
		      Help.render gfx;
		    gfx#sync ();
		    false, t -. refresh_interval
		  end
		else
		  world_changed, t
	    in
	    let state = 
	      { state with 
		  world_time = world_time;
		  world_changed = world_changed;
		  refresh_time = refresh_time } in
	      loop now state;
      | _ -> loop prev_time state
  in
    loop (Unix.gettimeofday ()) { cur_tanks = [Game.TankId 0];
				  world_time = 0.0;
				  world_changed = false;
				  refresh_time = 0.0;
				  input = None; 
				  guide = false; }

let main setup =
  let _ = Thread.create ticker () in
    ( match setup.setup_host with
	| None -> 
	    main_menu setup game_loop
	| Some server ->
	    let host = Unix.gethostbyname server in
	    let server = (Unix.ADDR_INET (host.Unix.h_addr_list.(0), 61753)) in
	    let client = Client.create server in
	      player_menu client game_loop );
    Printf.printf "Bye!\n%!";
    Perf.show_counters ()

let imain () =
  ()

let _ = 
  if !Sys.interactive then 
    imain ()
  else
    begin
      let use_meta = ref true in
      let connect_host = ref None in
      let invalid_params = ref false in
	Arg.parse [("-fs", Arg.Unit (fun () -> ()), "DON'T USE THIS SWITCH");
		   ("--nometa", Arg.Unit (fun () -> use_meta := false), "Disable meta-server support");
		   ("--server", Arg.String (fun s -> connect_host := Some s), "host Connect to Goba-server")]
	  (fun arg -> 
	     Printf.eprintf "Invalid parameter: %s\n" arg;
	     invalid_params := true )
	  "Game Of Bases";
	if !invalid_params then
	  ()
	else
	  begin
	    Random.self_init ();
	    Sdlwm.set_caption ~title:"Goba - Game Of Bases" ~icon:"Goba";
	    main { setup_meta = !use_meta;
		   setup_host = !connect_host }
	  end
    end
