TYPE_CONV_PATH "Goba"
open VecOps
open GobaUtils
open GobaMath

module B = Bitmaps

type t = { width	: int;
	   height	: int;
	   contents	: Tile.t array;
	   bitmaps	: ((int * int) * Sdlvideo.surface) option array;
	   tile_size	: int * int;
	 }

let black = (0, 0, 0)
let yellow = (255, 255, 200)
let white = (255, 255, 255)
let grey = (128, 128, 128)
let dark_grey = (64, 64, 64)
let green = (0, 255, 0)
let blue = (128, 128, 255)
let brown = (139, 69, 19)

let color_map (tile, info) = 
  let hp = info.Tile.hp in
    match tile with
      | Tile.Space -> white
      | Tile.Grass -> ipolate3 (f_scale 10.0 0.0 hp) green black
      | Tile.Water -> blue
      | Tile.Mountain -> ipolate3 (f_scale 50.0 0.0 hp) grey dark_grey
      | Tile.Wall -> white
      | Tile.Start _ -> failwith "No color for start place"

(* offset is used for multi-block tiles *)
let create_tile offset get =
  let mod_tile base =
    let xmod, ymod = fst offset mod 3, snd offset mod 3 in
      B.move (0, 20) (B.file (Printf.sprintf "%s-%d%d" base xmod ymod))
  in
  let grass_tile () =
    mod_tile "grass"
      (*    B.move (0, 0) (`Bitmap "grass")*)
  in
  let space_tile () = 
    mod_tile "space"
      (*B.empty*)
  in
  let (tile, info) = get offset in
  let masks =
    [(-1, 0), B.file "partial-mask-w";
     (0, -1), B.file "partial-mask-n";
     (1, 0), B.file "partial-mask-e";
     (0, 1), B.file "partial-mask-s"] 
  in
  let masks () = List.concat (
    List.map 
      (fun (ofs, mask) ->
	 match get (offset +| ofs) with
	   | (Tile.Grass, _) -> [mask]
	   | _ -> []
      )
      masks
  )
  in
  let description = 
    let color = color_map (tile, info) in
      match tile with 
	| Tile.Space -> 
	    let ms = masks () in
	      if ms = [] then
		space_tile ()
	      else
		B.combine (B.move (0, 20) (B.compose ms),
			   space_tile (),
			   grass_tile ())
	| Tile.Grass -> 
	    B.compose [grass_tile (); B.realpha (info.Tile.hp /. 10.0) (B.file "grass")];
	| Tile.Mountain ->
	    let ms = masks () in
	    let mountain = 
	      if (fst offset + snd offset) mod 2 = 0 then
		B.file "mountain1"
	      else
		B.file "mountain2"
	    in
	      if ms = [] then
		B.compose [space_tile ();
			   B.realpha (f_scale 0.0 50.0 info.Tile.hp) (B.move (0, 20) mountain)]
	      else
		B.compose [B.combine (B.move (0, 20) (B.compose ms),
				      space_tile (),
				      grass_tile ());
			   B.realpha (f_scale 0.0 50.0 info.Tile.hp) (B.move (0, 20) mountain)]
	| _ -> B.compose [space_tile (); B.move (0, 10) (B.char (Tile.char_of_tile tile, color))]
	    in
	      Bitmaps.bitmap description

let within_map map (x, y) =
  x >= 0 && y >= 0 && x < map.width && y < map.height

let at_full map (x, y) =
  if within_map map (x, y) then
    map.contents.(x + map.width * y)
  else
    Tile.tile_with_info Tile.Wall

let at map coord =
  fst (at_full map coord)

let bitmap_at map (x, y) =
  match map.bitmaps.(x + map.width * y) with
    | None -> 
	let t = create_tile (x, y) (at_full map) in
	  map.bitmaps.(x + map.width * y) <- Some t;
	  t
    | Some x -> x

let set map (x, y) value =
  map.contents.(x + map.width * y) <- value;
  map.bitmaps.(x + map.width * y) <- None

let damage map ((x, y) as at) damage_hp =
  if not (within_map map (x, y)) then
    0.0
  else
    let ch, info = at_full map at in
    let real_damage =
      match ch with
	| Tile.Grass 
	| Tile.Mountain -> damage_hp
	| _ -> 0.0
    in
    let { Tile.hp = hp } = info in
    let new_damage = max 0.0 (info.Tile.hp -. real_damage) in
    let delta = hp -. new_damage in
      set map at (ch, { Tile.hp = new_damage });
      delta
    
let tile_at map map_at = 
  if not (within_map map map_at) then
    Bitmaps.bitmap (B.move (0, 10) (B.char ('#', white)))
  else
    bitmap_at map map_at
	      
let create 
    { MapData.size = width, height; 
      tile_size = tile_size; 
      data = contents }  =
    { width	= width;
      height	= height;
      contents  = contents;
      tile_size = tile_size;
      bitmaps	= Array.map (fun _ -> None) contents
    }

let block_cmask =
  lazy (let bitmap = Bitmaps.bitmap (B.block (20, 20, (255, 255, 255), 255)) in
	Bitmaps.collision_mask_of_surface (snd bitmap))

class map_object map map_at tile_size at (collidable : bool) (destructible : bool) (friction : float) = 
object
  method get_type : ObjectType.t = failwith "get_type not supported on map tiles"
  method get_state : float -> Sexplib.Sexp.t = failwith "Cannot get_state map_object"
  method set_state : Sexplib.Sexp.t -> float -> unit = failwith "Cannot set_state map_object"
  method self_owned = true
  method get_id = Object.MapId map_at
  method collision_area = 
    let ofs = (0.0, snd tile_size) in
      (at +|. ofs, at +|. ofs +|. tile_size), Lazy.force block_cmask
  method location = 
    let ofs = tile_size /|. Vector.v2 2.0 in
      at +|. ofs
  method z_location = 0.0
  method render (gfx : Gfx.t) (_:Object.render_flags) (_at : int * int) : (int * (unit -> unit)) list = 
    []
  method impact (world : Object.world) (damage_hp : float) : unit = 
    let got_damage = damage map map_at damage_hp in
      world#report_damage (Object.MapId map_at, got_damage);
    
  method collide (world : Object.world) = collidable

  method speed = 0.0, 0.0
  method set_speed (speed : float * float) = ()
  method collidable = collidable
  method destructible = destructible

  method friction = friction

  method cause_damage (_:float) = ()

  method is_alive = true

  method change = ()

  method tick (_:float) (_:Object.world) = ()

    (* temporary hack *)
  method set_self_owned (_:bool) = ()

  method set_location (at' : float * float) : unit = failwith "attempted to set a map tile coordinate"

  method changed_since (_:Object.serial) = false

  method test (_:Object.world) = ()

  method recharge = false

  method finish = ()
end

let real_coords map at =
  Vector.float2 (at *| map.tile_size)

let map_coords map at =
  at /|. Vector.float2 map.tile_size

let map_int_coords map at =
  at /| map.tile_size,
  Vector.mod2 at map.tile_size
    
let objects_near map ((x1, y1), (x2, y2)) : Object.t list =
  let objs = ref [] in
  let x1, y1 = map_coords map (x1, y1) in
  let x2, y2 = map_coords map (x2, y2) in
    for yc = (int_of_float (y1 -. 1.5)) to (int_of_float (y2 +. 0.0)) do
      for xc = (int_of_float (x1 -. 1.5)) to (int_of_float (x2 +. 0.0)) do
	let ch, info = 
	  if xc >= 0 && yc >= 0 && xc < map.width && yc < map.height then
	    at_full map(xc, yc)
	  else
	    Tile.tile_with_info (Tile.Wall)
	in
	let friction =
	  let hp = info.Tile.hp in
	  match ch with
	    | Tile.Grass when hp >= 5.0 -> 0.37
	    | Tile.Grass when hp >= 1.0 -> 0.15
	    | Tile.Grass -> 0.0
	    | Tile.Mountain when hp > 10.0 -> 0.4
	    | Tile.Mountain -> 0.0
	    | _ -> 0.0
	in
	let collidable, destructible =
	  let hp = info.Tile.hp in
	  match ch with
	    | Tile.Mountain when hp > 25.0 -> true, true
	    | Tile.Mountain -> false, true
	    | Tile.Wall -> true, true
	    | Tile.Grass -> false, true
	    | _ -> false, false
	in
	  if info.Tile.hp > 0.0 && ch <> Tile.Space then
	    let obj = new map_object map (xc, yc) (Vector.float2 map.tile_size) (real_coords map (xc, yc)) collidable destructible friction in
	      objs := (obj :> Object.t) :: !objs
      done
    done;
    !objs

let ord_partition f l =
  let rec aux tail head =
    match tail with
      | x::xs when f x ->
	  aux xs (x::head)
      | _ ->
	  List.rev head, tail
  in
    aux l []

type ('mask, 'b) renderable =
    < collision_area : ((float * float) * (float * float)) * 'mask;
      z_location : float; 
      location : float * float;
      render : Gfx.t -> Object.render_flags -> int * int -> (int * (unit -> unit)) list;
      .. > as 'b 

let render_map map (gfx:Gfx.t) (flags:Object.render_flags) display_at (offset, size) 
    (objects: ('mask, 'a) renderable list) 
    (overlay : 'b list) =
  gfx#with_clip_rect 
    (Sdlvideo.rect 
       (fst display_at) (snd display_at) 
       (int_of_float (fst size)) (int_of_float (snd size)))
    (fun () ->
       let offset = Vector.round2 offset in
       let map_offset = map_int_coords map offset in
       let map_size = Vector.int2 size /| map.tile_size in
	 (*       let offset_frac = 
		  let frac v = fst (modf v) in
		  frac (fst map_offset), frac (snd map_offset)
		  in*)
       let depth (y, z) = y -. z in
       let objs =
	 List.sort
	   (fun (y1, _) (y2, _) -> compare y1 y2)
	   (List.concat
	      (List.map (fun o -> 
			   (*let (((_, _), (_, y)), _) = o#collision_area in *)
			   (* why add z_location here? we want to
			      consider this object to be more south than
			      it actually is, for it to be rendered later
			   *)
			   let y = snd (o#location) in
			     List.map
			       (fun (z_ofs, render) ->
				  (depth (y, o#z_location +. float z_ofs), render)
			       )
			       (o#render gfx flags
				  (Vector.round2 o#location -| offset +| display_at))
			)
		 objects)
	   )
       in
       let map_objs =
	 GobaUtils.foldr2d 
	   (fun (x, y) objs ->
	      let at_y = float (snd offset + y * snd map.tile_size - snd (snd map_offset)) in
	      let render () =
		let coord = (x, y) in
		let at = Vector.round2 (Vector.float2 coord *|. Vector.float2 map.tile_size) -| snd map_offset in
		let map_at = coord +| fst map_offset in
		let (surface_offset, surface) = tile_at map map_at in
		  gfx#blit (at +| surface_offset +| display_at) surface
	      in
		(depth (at_y, 0.0), render)::objs
	   )
	   []
	   ((-2, -2), (fst map_size + 2, snd map_size + 3))
       in
       let render_objs objs =
	 List.iter (fun (_, obj) -> 
		      obj ()
		   ) objs
       in
	 render_objs (List.merge (fun (y1, _) (y2, _) -> compare y1 y2) objs map_objs);
	 List.iter 
	   (fun o ->
	      List.iter 
		(fun (_, render) -> render ())
		(o#render (gfx :> Gfx.t) flags
		   (Vector.round2 o#location -| offset +| display_at))
	   ) overlay;
    )
