TYPE_CONV_PATH "Goba"
module type ACCURACY =
  sig
    type t
    val add : t -> t -> t
    val sub : t -> t -> t
    val mul : t -> t -> t
    val zero : t
    val eq : t -> t -> bool
    val ge : t -> t -> bool
    val gt : t -> t -> bool
    val to_float : t -> float 
    val of_float : float -> t
    val print : t -> out_channel -> unit
  end

module FloatAccuracy =
  struct
    type t = float
    let add = (+.)
    let sub = (-.)
    let mul = ( *. )
    let eq = (=)
    let ge = (>=)
    let gt = (>)
    let zero = 0.0
    let to_float v = v
    let of_float v = v
    let print v ch = Printf.fprintf ch "%.1f" v
  end

module IntAccuracy =
  struct
    type t = int
    let add = (+)
    let sub = (-)
    let mul = ( * )
    let eq = (=)
    let ge = (>=)
    let gt = (>)
    let zero = 0
    let to_float v = float_of_int v
    let of_float v = int_of_float v
    let print v ch = Printf.fprintf ch "%d" v
  end

module Utils =
  struct
    let rec take n l = 
      match n, l with 
	| _, []	-> []
	| 0, _	-> []
	| n, hd::tl -> hd::take (n - 1) tl

    let flip a =
      fun x y ->
	a y x

    let (@@) a b = a b
    let (@.) a b = fun c -> a (b c)

    (* maximum, minimum *)
    let imum test data =
      List.fold_left ( fun best current -> if test best current then best else current ) (List.hd data) (List.tl data)

    let imums test data =
      imum test data, imum (fun a b -> not (test a b)) data

    let numerate xs =
      let rec f n xs =
	match xs with
	  | [] -> []
	  | x::xs -> (n,x) :: f (n + 1) xs
      in
	f 0 xs

    let swap (a, b) = (b, a)

    let rec takeWhile f l =
      match l with
	| x::xs when f x ->
	    x::takeWhile f xs
	| _ -> []

    let rec dropWhile f l =
      match l with
	| x::xs when f x ->
	    dropWhile f xs
	| _ -> l

    let splitWhen f l =
      takeWhile (not @. f) l, dropWhile (not @. f) l
  end

module Ops =
  functor (Acc : ACCURACY) ->
    struct
      (*type t = Acc.t*)
	  
      type point = (Acc.t * Acc.t)

      type side = Left | Right

      let ( +.. ) = Acc.add

      let ( -.. ) = Acc.sub

      let ( *.. ) = Acc.mul

      let ( =.. ) = Acc.eq

      let ( >=.. ) = Acc.ge

      let ( <.. ) a b = not (Acc.ge a b)

      let ( >.. ) = Acc.gt

      let ( -| ) (x1, y1) (x2, y2) =
	(x1 -.. x2, y1 -.. y2)

      let ( +| ) (x1, y1) (x2, y2) =
	(x1 +.. x2, y1 +.. y2)

    end

module Convexifier =
  functor (Acc : ACCURACY) ->
    struct
      module O = Ops(Acc)

      open O
      open Utils

      let side (v1,v2) p =
	let v = v2 -| v1 in
	let p = p -| v1 in
	  if fst v *.. snd p -.. fst p *.. snd v >=.. Acc.zero then
	    Left
	  else
	    Right

      let floatise (a, b) =
	Acc.to_float a, Acc.to_float b

      let unitise (a, b) =
	let length = sqrt (Acc.to_float (a *.. a +.. b *.. b)) in
	  Acc.to_float a /. length, Acc.to_float b /. length

      let o'p (x, y) chan =
	Printf.fprintf chan " (%t,%t) " (Acc.print x) (Acc.print y)

      let o'ps ps chan =
	Printf.fprintf chan "[";
	List.iter (flip o'p chan) ps;
	Printf.fprintf chan "]"

      let is_convex (points : point list) : bool =
	let rec check points =
	  match points with
	    | p1::((p2::p3::_) as ps) when side (p1, p3) p2 = Left ->
		false
	    | p::ps ->
		check ps
	    | _ -> true
	in
	let (points, points') = points, List.rev points in
	  check points && 
	    check (List.rev (take 2 points') @ take 1 points) &&
	    check (List.rev (take 1 points') @ take 2 points)

(*
      let giftwrap (points : point list) : point list =
	let p = List.fold_left (fun p1 p2 -> if snd p2 <. snd p1 then p2 else p1) (List.hd points) (List.tl points) in
  []*)

      let point_distance ((a : point), (b : point)) (p : point) : float =
	let p1 = unitise (b -| a) in
	let p2 = floatise (p -| a) in
	  fst p1 *. fst p2 +. snd p1 *. snd p2 

      let left_side segment point =
	side segment point = Left

      let chord points =
	let a = imum ( fun (_, (x1, _)) (_, (x2, _)) -> x1 <.. x2 ) points in
	let b = imum ( fun (_, (x1, _)) (_, (x2, _)) -> x1 >.. x2 ) points in
	  (a, b)

      let rec do_quick_hull (points : (int * point) list) (a, b) =
	let segment = match (a, b) with (_, a), (_, b) -> a, b in
	let distance = point_distance segment in
	let most_far = ( fun (_, a) (_, b) -> distance a < distance b ) in
	let points = List.filter ( fun (x, _) -> x <> fst a & x <> fst b ) points in
	let left = List.filter ( left_side segment @. snd ) points in
	  if left <> [] then
	    let left_far = imum most_far left in
	      [left_far] @
		do_quick_hull left (a, left_far) @
		do_quick_hull left (left_far, b)
	  else 
	    ( [] )
	      
      let quickhull (points : point list) : point list =
	let points = numerate points in
	  if points = [] then
	    []
	  else
	    let (a, b) = chord points in
	    let segment = match (a, b) with (_, a), (_, b) -> a, b in
	    let distance p = point_distance segment (snd p) in
	    let most_far = ( fun a b -> distance a < distance b ) in
	    let points = List.filter ( fun (x, _) -> x <> fst a & x <> fst b ) points in
	    let (left, right) = List.partition ( left_side segment @. snd ) points in
	    let res = [a] @ (do_quick_hull left (a, b)) @ [b] @ (do_quick_hull right (b, a))
	    in
	      List.map snd res

      let pi = atan (1.0) *. 4.0

      let angle (vx, vy) =
	let t = if (vx > vy) then abs_float ( atan (vy /. vx) ) else pi /. 2.0 -. abs_float ( atan (vx /. vy) ) in
	  match (vx >= 0.0, vy >= 0.0) with
	    | (false, false) -> pi +. t
	    | (true,  false) -> 2.0 *. pi -. t
	    | (false, true)  -> pi -. t
	    | (true,  true)  -> t

      let graham_scan (points : point list) : point list =
	let points = numerate points in
	let (pivot_n, pivot) = imum ( fun (_, (_, y1)) (_, (_, y2)) -> y1 <.. y2 ) points in

	let rec scan behind ahead =
	  match ahead with
	    | p1::p2::p3::ps when side (p1, p3) p2 = Left ->
		if behind <> [] then
		  scan (List.tl behind) (List.hd behind::p1::p3::ps)
		else
		  scan behind (p1::p3::ps)
	    | p1::p2::[] when side (p1, pivot) p2 = Left ->
		if behind <> [] then
		  scan (List.tl behind) (List.hd behind::p1::[])
		else
		  scan behind (p1::[])
	    | p1::ps ->
		scan (p1::behind) ps
	    | [] -> List.rev behind
	in

	let points = List.filter ( fun (n, _) -> n <> pivot_n ) points in
	let points = List.map ( fun (n, ((x, y) as p)) -> 
				  let d = (x, y) -| pivot in
				    angle (floatise d), p ) points in
	let points = List.sort ( fun (a1, _) (a2, _) -> compare a1 a2 ) points in
	  (*Printf.printf "Pivot : %t  Points : %t\n%!" 
	    (o'p pivot) 
	    (fun ch -> List.iter (fun (ang, p) -> Printf.fprintf ch " %.2f:%t " ang (o'p p)) points);*)
	let points = List.map snd points in
	  scan [] (pivot::points)

      let convexify = graham_scan
    end

module Coverage =
  functor (Acc : ACCURACY) ->
    struct
      module O = Ops(Acc)
      module H = Convexifier(Acc)
      open O
      open Utils

      type segment = ((Acc.t * Acc.t) * (Acc.t * Acc.t))

      let segments ps =
	List.combine ps (List.tl ps @ [List.hd ps])

      let right_side ((_, y1), (_, y2)) =
	y1 <.. y2

      let left_side ((_, y1), (_, y2)) =
	y2 <.. y1

      let sides ps =
	let segs = segments @@ H.graham_scan ps in
	let left = List.map swap @@ List.rev @@ List.filter left_side segs in
	let right = List.filter right_side segs in
	  left, right

      let advanceSide scan l =
	splitWhen (fun ((_, y1), (_, y2)) -> y2 >=.. scan) l

      let point_of_seg ((x1, y1), (x2, y2)) y =
	let slope = (Acc.to_float (x2 -.. x1)) /. (Acc.to_float (y2 -.. y1)) in
	  Acc.of_float (Acc.to_float x1 +. (Acc.to_float y -. Acc.to_float y1) *. slope), y

      let combinableRects threshold ((r1'x1, r1'y1), (r1'x2, r1'y2)) ((r2'x1, r2'y1), (r2'x2, r2'y2)) =
	let fthresh = Acc.to_float threshold in
	  abs_float (Acc.to_float (r1'x1 -.. r2'x1)) < fthresh /. 2.0 &&
	    abs_float (Acc.to_float (r1'x2 -.. r2'x2)) < fthresh /. 2.0 &&
	    abs_float (Acc.to_float (r1'x1 -.. r2'x1)) +. abs_float (Acc.to_float (r1'x2 -.. r2'x2)) < fthresh

      let combineRects ((r1'x1, r1'y1), (r1'x2, r1'y2)) ((r2'x1, r2'y1), (r2'x2, r2'y2)) =
	((imum (<..) [r1'x1; r2'x1], imum (<..) [r1'y1; r2'y1]),
	 (imum (>..) [r1'x2; r2'x2], imum (>..) [r1'y2; r1'y2]))

      let o'seg (p1, p2) chan =
	Printf.fprintf chan " (%t-%t) " (H.o'p p1) (H.o'p p2)
	  
      let o'segs ss chan =
	List.iter ( fun p -> o'seg p chan ) ss

      let rect_coverage slice ps =
	let left, right = sides ps in
	let (_, min'y), (_, max'y) = imums (fun (_, y1) (_, y2) -> y1 < y2) ps in

	let rec traverse scan (left : segment list) (right : segment list) rects prevRange =
	  let leftSkip, leftAhead = advanceSide scan left in
	  let rightSkip, rightAhead = advanceSide scan right in
	  let prevScan = scan -.. slice in
	  let nextScan = scan +.. slice in
	  let to_xs = List.concat @. List.map (fun ((_, _), (x2, _)) -> [x2]) in
	    (*Printf.printf "\n\n";*)
	    if (leftAhead <> [] &&
		  rightAhead <> []) then
	      let leftPoint = point_of_seg (List.hd leftAhead) scan in
	      let rightPoint = point_of_seg (List.hd rightAhead) scan in
	      let minLeft = imum (<..) @@ fst leftPoint :: to_xs leftSkip in
	      let maxRight = imum (>..) @@ fst rightPoint :: to_xs rightSkip in
		(*Printf.printf "scan = %f - %f (-> %f)\n" (Acc.to_float prevScan) (Acc.to_float scan) (Acc.to_float nextScan);
		  Printf.printf "left:%t/%t right:%t/%t\n" 
		  (o'segs leftSkip) (o'segs leftAhead) 
		  (o'segs rightSkip) (o'segs rightAhead);
		  Printf.printf "Skipped points (left): %t\n" 
		    (fun ch -> List.iter (fun f -> Printf.printf " %f" (Acc.to_float f)) (to_xs leftSkip));
		  Printf.printf "Skipped points (right): %t\n"
		    (fun ch -> List.iter (fun f -> Printf.printf " %f" (Acc.to_float f)) (to_xs rightSkip));
		  Printf.printf "range %f - %f\n" (Acc.to_float minLeft) (Acc.to_float maxRight);
		  Printf.printf "%!";*)
	      let rects = 
		match prevRange with
		  | None -> rects
		  | Some (minLeft', maxRight') ->
		      let rect = ((min minLeft minLeft', prevScan), (max maxRight maxRight', scan))
		      in 
			match rects with
			  | rect'::rects' when combinableRects slice rect rect' ->
			      combineRects rect rect'::rects'
			  | _ -> rect::rects
	      in traverse nextScan leftAhead rightAhead rects (Some (fst leftPoint, fst rightPoint))
	    else
	      let rects =
		match prevRange with
		  | None -> rects
		  | Some (minLeft', maxRight') ->
		      let minLeft = imum (<..) @@ minLeft' :: to_xs leftSkip in
		      let maxRight = imum (>..) @@ maxRight' :: to_xs rightSkip in
		      let rect = (minLeft, prevScan), (maxRight, (min scan max'y)) in
			(*
			  Printf.printf "Skipped points (left): %t\n" 
			  (fun ch -> List.iter (fun f -> Printf.printf " %f" f) (to_xs leftSkip));
			  Printf.printf "Skipped points (right): %t\n" 
			  (fun ch -> List.iter (fun f -> Printf.printf " %f" f) (to_xs rightSkip));
			  Printf.printf "maxRight = %f\n" maxRight;
			  Printf.printf "%!";*)
			(*marker (10.0, 20.0) (snd rect) 10;*)
			rect::rects
	      in List.rev rects
	in    
	  traverse min'y left right [] None
      
    end

module Int = Convexifier (IntAccuracy)

module Float = Convexifier (FloatAccuracy)
