open VecOps

type 'a entry = 
    { label : string;
      value : 'a; }

(** menu_size -> screen_size -> location -> screen_coordinate *)
type justify = 
  ?menu_size : (int * int) -> 
  screen_size : int * int ->
  size : int * int ->
  at : int * int ->
  int * int

type area = ((int * int) * (int * int))

type 'a t = { entries		: 'a entry list;
	      mutable selected	: int;
	      mutable time	: float;
	      justify		: screen_size : int * int ->
					 size : int * int ->
					 at : int * int ->
					 int * int;
	      font		: Sdlttf.font;
	    }

let entry label value =
  { label = label;
    value = value }

let foldl_entries : 
  f:('a -> int -> Sdlvideo.surface -> 'a) ->
  font:Sdlttf.font ->
  v0:'a ->
  time:float -> 
  selected:int -> 
  entries:'b entry list -> 'a 
  =
  
  fun ~f ~font ~v0 ~time ~selected ~entries ->
    let (_, _, v) = 
      List.fold_left
	(fun (y, nth, v) { label = label } ->
	   let color = 
	     if nth = selected then
	       (255, min 255 (max 0 (200 + int_of_float (50.0 *. sin (2.0 *. time)))), 255)
	     else
	       (100, 200, 200)
	   in
	   let text = 
	     Sdlvideo.display_format ~alpha:true 
	       (Sdlttf.render_text_blended font label color)
	   in
	   let v = f v y text in
	   let (_, h) = Gfx.surface_dims text in
	     (y + h + 20, nth + 1, v)
	)
	(0, 0, v0)
	entries
    in
      v

let center ?menu_size:(menu_w, menu_h=(0, 0)) ~screen_size:(screen_w, screen_h) ~size:(w, h) ~at:(x, y) =
  Printf.printf "Screen size: %d,%d menu size:%d,%d\n%!" screen_w screen_h menu_w menu_h;
  ((screen_w - w) / 2, (screen_h - menu_h - h) / 2 + y)

let align_south =
  fun 
    ?(marginal=50)
    ?menu_size:(menu_w, menu_h=(0, 0))
    ~screen_size:(screen_w, screen_h) 
    ~size:(w, h)
    ~at:(x, y) ->
    ((screen_w - w) / 2, screen_h - marginal - menu_h + y)

let entries_area (justify:justify) screen_size font entries =
  GobaUtils.deoption (
    foldl_entries 
      ~f:(fun area y text ->
	    let size = Gfx.surface_dims text in
	    let (x0, y0) = justify ?menu_size:None ~screen_size ~size ~at:(0, y) in
	    let (x1, y1) = (x0, y0) +| size in
	      match area with
		| None -> Some ((x0, y0), (x1, y1))
		| Some ((x0', y0'), (x1', y1')) ->
		    Some ((min x0 x0', min y0 y0'),
			  (max x1 x1', max y1 y1'))
	 )
      ~v0:None
      ~font
      ~time:0.0
      ~selected:0
      ~entries
  )

let create ?(justify:justify=center) ?(font=Gfx.get_font ~size:40 ()) ?(selected=0) entries =
  let area = 
    let (a, b) = entries_area justify (0, 0) font entries in
      b -| a
  in
    { entries = entries;
      selected = max 0 (min selected (List.length entries - 1));
      time = 0.0;
      font = font;
      justify = justify ~menu_size:area; }

let currently_selected t = t.selected

let choose_next t =
  t.selected <- min (t.selected + 1) (List.length t.entries - 1)

let choose_prev t =
  t.selected <- max (t.selected - 1) 0

let choose t =
  (List.nth t.entries t.selected).value

let tick t delta =
  t.time <- t.time +. delta

let render t (gfx:Gfx.t) =
  let screen_size = Gfx.surface_dims gfx#get_surface in
    foldl_entries
      ~f:(fun () y text ->
	    let size = Gfx.surface_dims text in
	      gfx#blit (t.justify ~screen_size ~size ~at:(0, y)) text;
	 )
      ~v0:()
      ~font:t.font
      ~time:t.time
      ~selected:t.selected
      ~entries:t.entries
