open GobaMath
open GobaOps
open VecOps

type waypoint = (float * float)

let step_size = 20.0
let angle_step = pi /. 7.0

(*
  ..........
  ........a.
  ...#.../..
  ..###./...
  ..###/-...
  ...###|-..
  ....###|..
  .....#/...
  .b---/....
  ..........

  ..........
  ........a.
  ...#.../..
  ..###..|..
  ..###..|..
  ...###.|..
  ....###|..
  .....#/...
  .b---/....
  ..........

*)

let mask = CollisionMask.circle_loc_mask 15

class mask at =
object
  inherit Collision.t
  method collision_area = Collision.carea_of_mask at mask
end 

let mask_size = (40.0, 40.0)

let cont_opt v f' =
  match v with
    | None -> f' ()
    | Some _ -> v

let optimize_path (world:Object.world) path =
  let obstructions_between a b =
    let delta = b -|. a in
    let step = Vector.unit delta *|. Vector.v2 step_size in
    let num_steps = int_of_float (Vector.abs2 delta /. step_size -. 1.0) in
      GobaUtils.until_result
	(fun (at, steps_left) ->
	   if steps_left > 0 then
	     match world#collision (new mask at) (fun obj -> obj#collidable) with
	       | None -> GobaUtils.NoResult (at +|. step, steps_left - 1)
	       | Some (_, objs) -> 
		   GobaUtils.SomeResult true
	   else
	     GobaUtils.SomeResult false
	)
	(a +|. step, num_steps)
  in
  let rec aux path prev =
    match path with
      | x::_::y::rest when not (obstructions_between x y) ->
	  (match prev with
	     | p1::ps -> aux (p1::x::y::rest) ps
	     | [] -> aux (x::y::rest) prev)
      | x::rest ->
	  aux rest (x::prev)
      | [] -> List.rev prev
  in
    aux path []

let plan (world:Object.world) ~src ~dst =
  let backtracks_left = ref 500 in
  let rec sweep_angles at ang dead_angle step_size aux : 'a list option =
    let rec sweep_aux ofs =
      let check_angle ang : 'a list option =
	if match dead_angle with 
	  | None -> true 
	  | Some dead_angle when 
	      let d = abs_float (angle_delta dead_angle ang) in
		(*Printf.printf "angle=%f dead_angle=%f delta=%f limit=%f\n"
		  ang dead_angle d (pi /. 4.0);*)
		d < pi /. 4.0 ->
	      false
	  | _-> true
	then
	  (*Printf.printf "angle=%f\n%!" ang;*)
	  let at' = at +|. Vector.v2 step_size *|. (cos ang, sin ang) in
	    aux (Some (ang +. pi)) at' dst (fun at -> None)
	else
	  None
      in
	if ofs < pi +. 0.0005 then
	  begin
	    let cont () = sweep_aux (ofs +. angle_step) in
	      cont_opt (check_angle (ang +. ofs)) @@ fun () ->
		if ofs > 0.0 then
		  cont_opt (check_angle (ang -. ofs)) @@ cont
		else
		  cont ()
	  end
	else
	  None
    in
      if !backtracks_left <= 0 then
	None
      else
	sweep_aux 0.0
  in
  let rec plan_aux prev_angle at dst decide_collision : 'a list option =
    if Vector.abs2 (dst -|. at) < step_size then
      Some [at]
    else 
      match world#collision (new mask at) (fun obj -> obj#collidable) with
	| Some _ ->
	    (*Printf.printf "Dead end at %s\n%!" (Vector.f_to_string at);
	      None*)
	    decide_collision at
	| None -> (* ok, it's free here *)
	    decr backtracks_left;
	    let ang = Vector.angle (dst -|. at) in
	      (*Printf.printf "at=%s\n%!" (Vector.f_to_string at);*)
	      match sweep_angles at ang prev_angle step_size plan_aux with
		| None -> None
		| Some x -> Some (at::x)
  in
  let path = 
    plan_aux None src dst
      (fun _at ->
	 let ang = (Vector.angle (dst -|. src)) in
	   sweep_angles src ang (Some ang) (step_size) plan_aux)
  in
  let path = 
    match path with
      | None -> None
      | Some path -> Some (optimize_path world path)
  in
    path
