TYPE_CONV_PATH "Goba"

open VecOps
open GobaMath
open GobaOps

let firing_period = 0.2
let mine_period = 1.0
let max_speed = 150.0
let max_reverse_speed = -.50.0
let opposition = 1.0 (* wind force *)
let friction = 0.60

let bitmap_offset = (-60, -60)
let bitmap_size = (120, 120)

let max_hp = 100.0
let max_ammo = 100
let max_mines = 4
let max_building_points = 10

let ident_font = lazy (Gfx.font 20)

let round_direction ang steps =
  let ofs = pi2 /. float steps /. 2. in 
  let (ang, angle_mul) = 
    if ang < 0.0 then (-. ang, -. 1.0)
    else (ang, 1.0)
  in
    angle_mul *. (ang -. mod_float (ang +. ofs) (pi2 /. float steps) +. ofs)

let to_zero v v' =
  let v' = 
    if v < 0.0 then
      max 0.0 v'
    else
      min 0.0 v'
  in
    if abs_float v' > abs_float v then
      v
    else
      v'

let frames color = 
  Array.init 36
    (fun n ->
       lazy (
	 let module B = Bitmaps in
	 let nc = B.no_cache in
	 let bm = B.bitmap @@ B.move bitmap_offset
			      @@ B.compose
				 [B.colored color @@
				    B.auto_crop @@
				    nc @@ B.file @@ Printf.sprintf "tank_value-%04d" (n + 1);
				  B.auto_crop @@
				    nc @@ B.file @@ Printf.sprintf "tank_color-%04d" (n + 1);
				 ] in
	 let mask = B.mask_with_offset B.collision_mask_of_surface2 @@
	   B.bitmap @@ B.move bitmap_offset @@
	   B.auto_crop @@ B.file @@ Printf.sprintf "tank_mask-%04d" (n + 1)
	 in
	   (bm, mask))
    )

let collision_sound =
  let sample = Sfx.Sample "collision" in
  let channel = Sfx.alloc_channel () in
    fun (world : Object.world) ->
      let ch = world#channel channel in
	if not ch#playing then ch#play sample

type init = { init_at : float * float;
	      init_color : int * int * int;
	      init_direction : float;
	      init_hp : float;
	      init_ammo : int;
	      init_mines : int;
	      init_building_points : int; }
with sexp

let defaults =
  { init_at = (0.0, 0.0);
    init_color = (255, 255, 255);
    init_direction = 0.0;
    init_hp = max_hp;
    init_ammo = max_ammo;
    init_mines = max_mines;
    init_building_points = max_building_points }

class t id init0 =
  (*let bitmap = lazy (Bitmaps.bitmap (Bitmaps.block (11, 18, constr.Constr.tank_color), 255)) in*)
  (*let cmask = lazy (Bitmaps.collision_mask_of_surface (snd (Lazy.force bitmap))) in*)

  let init = Object.default_with init0 init_of_sexp in

object (self)
  inherit Object.t id as super

  val frames = frames init.init_color

  val mutable at = init.init_at
  val mutable motor_speed = 0.0
  val mutable speed = (0.0, 0.0)
  val mutable firing_time = firing_period		(* how long since the last time we fired? *)
  val mutable mine_time = mine_period (* how long since we last dropped a mine? *)

  val mutable ident_surface = None
    
  val mutable hp = init.init_hp
  val mutable ammo = init.init_ammo
  val mutable mines = init.init_mines
  val mutable building_points = init.init_building_points
  val mutable direction = 0.0

  val mutable prev_map_obj_ids = []

  val mutable damage_counter = 0.0

  val mutable goal = None

  val mutable path : PathPlanner.waypoint list option = None

  method get_type = ObjectType.Tank

  method get_state delay =
    Object.sexp_of_state sexp_of_init
      { Object.state_super = super#get_state delay;
	state_init = { init_at = init.init_at;
		       init_color = init.init_color;
		       init_direction = direction;
		       init_hp = hp;
		       init_ammo = ammo;
		       init_mines = mines;
		       init_building_points = building_points; } }

  method set_state state delay =
    let state = Object.state_of_sexp init_of_sexp state in
      super#set_state state.Object.state_super delay;
      let init = state.Object.state_init in
	direction <- init.init_direction;
	hp <- init.init_hp;
	ammo <- init.init_ammo;
	mines <- init.init_mines;
	building_points <- init.init_building_points;

  method get_direction = direction

  method z_location = 0.0

  method set_ident str = 
    let surface = Sdlttf.render_text_blended (Lazy.force ident_font) str (128, 255, 128) in
      ident_surface <- Some surface

  method set_location at' = at <- at'
  method set_speed speed' = 
    speed <- speed';
    (* not quite correct *)
(*    if Vector.abs2 speed > 0.01 then
      direction <- atan2 (snd speed) (fst speed)*)

  method turn adjust =
    direction <- direction +. adjust;
    self#change

  method cur_frame =
    Lazy.force (
      frames.(GobaMath.frame_of_radian direction)
    )

  method get_damage_counter = damage_counter

  method location = at
  method speed = speed
  method render gfx flags at_display = 
    [(0, fun () ->
	Bitmaps.blit gfx at_display (fst self#cur_frame);
	(match path with 
	   | None -> () 
	   | Some ps -> 
	       List.iter 
		 (fun p -> gfx#rectangle (255, 255, 255) (Vector.int2 (p -|. at) +| at_display) (3, 3))
		 ps
	);
	if List.exists 
	  (function 
	     | `Labels -> true
	     | `LabelsExcept oids when not (List.exists ((=) self#get_id) oids) -> true
	     | _ -> false
	  )
	  flags then 
	    match ident_surface with
	      | None -> ()
	      | Some surface ->
		  (*	    let size = Gfx.surface_dims surface in*)
		  (* doesn't work ??? *)
		  (*	    let pos = (at -| size /| (2, 2) +| ofs) in*)
		  let pos = (at_display +| (*Vector.int2 bitmap_offset +| *)(13, -15)) in
		    (*	      Printf.printf "size = %d,%d\n%!" (fst size) (snd size);
			      Printf.printf "pos = %d,%d\n%!" (fst pos) (snd pos);*)
		    gfx#blit pos surface
     )]

  method get_hp = hp

  method accelerate adjust = 
    motor_speed <- max max_reverse_speed (min max_speed (motor_speed +. adjust));
    self#change

  method turret_offset (angle : float) =
    (*Vector.float2 bitmap_size /|. Vector.v2 2.0*)
    (0.0, 0.0)

  method private adjust_location adj = at <- at +|. adj

  method private stop_axis vec = 
    let axis = Vector.unit vec in
    let power_fraction = 
      let direction' = round_direction direction 36 in
	abs_float (Vector.dot (cos direction', sin direction') axis)
    in
      (*Printf.printf "Power fraction: %f\n%!" power_fraction;*)
      motor_speed <- motor_speed *. power_fraction;
      speed <- speed -|. (axis *|. Vector.dup (Vector.dot speed axis));
      self#change
      
  method decelerate (amount : float) = 
(*    let abs = Vector.abs2 speed in
      if abs > amount then
	begin
	  speed <- speed -|. Vector.unit speed *|. Vector.dup amount;
	  self#change;
	end
      else
	begin
	  if speed <> (0.0, 0.0) then
	    self#change;
	  speed <- (0.0, 0.0)
	end*) ()

  method is_alive = hp > 0.0

  method tick delta world =
    let ratio = 0.01 in
    let direction' = round_direction direction 36 in
    let motor_vec = Vector.v2 (1.0 -. ratio ** delta) *|. (cos direction', sin direction') *|. Vector.v2 motor_speed in
      speed <- Vector.v2 (ratio ** delta) *|. speed +|. motor_vec;
      (*      if Vector.abs2 speed > 0.0 then
	      Printf.printf "%f %f\n%!" (Vector.abs2 speed) motor_speed;*)
      motor_speed <- 0.3 ** delta *. motor_speed;
      at <- at +|. Vector.v2 delta *|. speed;
      firing_time <- firing_time +. delta;
      damage_counter <- min 1.0 (max 0.0 (damage_counter *. (0.5 ** delta)-. delta));
      mine_time <- mine_time +. delta;
      let map_objs = (world#env (self :> Collision.t) (fun _ -> true)) in
      let map_friction = 
	List.fold_left 
	  (fun friction o -> max friction o#friction)
	  0.0
	  map_objs 
      in
	if self#self_owned then
	  begin
	    (* if we enter a new tile, cause a little damage to it *)
            List.iter 
              (fun o ->
		 if not (List.exists ((=) o#get_id) prev_map_obj_ids) then
		   o#impact world 1.0
              )
              map_objs;
	    (* also cause damage when we turn *)
	    let dir_change_damage = (0.5 *. motor_speed *. Vector.abs2 (Vector.unit motor_vec -|. Vector.unit speed)) *. delta in
	      if dir_change_damage > 0.0 then
		List.iter (fun o -> o#impact world dir_change_damage) map_objs;
	  end;
	prev_map_obj_ids <- List.map (fun o -> o#get_id) map_objs;
	speed <- 
	  (Vector.dup ((max 0.0 (1.0 -. friction -. map_friction)) ** delta) *|. 
	   let abs = Vector.abs2 speed -. opposition *. delta in
	     if abs <= opposition *. delta then
	       0.0, 0.0
	     else
	       (abs, abs) *|. Vector.unit speed);
	if Vector.abs2 speed > 0.0 then
	  self#change;
	if self#self_owned then
	  begin
	    self#autoguide world delta;
	    match world#collision (self :> Collision.t) (fun o -> o#collidable) with
	      | Some (x, objs) ->
		  let objs = List.filter (fun o -> o#collide world) objs in
		    if objs <> [] then
		      begin
			collision_sound world;
(*			Printf.printf "Adjust vector: %s %f %f\n%!" (Vector.f_to_string x) (Vector.abs2 x) (Vector.angle x);*)
			self#adjust_location x; (*(Vector.unit x *|. Vector.v2 (Vector.abs2 x +. 5.0)); *)
			self#stop_axis x
		      end
	      | None -> ()
	  end


  val mutable ag_k'p = 0.0
  val mutable ag_k'i = 0.0
  val mutable ag_k'd = 10.0

  val mutable ag_i = 0.0
  val mutable ag_d = 0.0
  val mutable ag_error = 0.0

  method get_goal =
    match goal with
      | None -> at
      | Some at -> at

  method private autoguide_break delta =
    let (p, i, d) = (ag_k'p, ag_k'i, ag_k'd) in
    let want_value = 0.0 in
    let value = Vector.dot speed (cos direction, sin direction) in
    let error = GobaMath.abs_limit 0.2 (want_value -. value) in
      ag_i <- ag_i +. error;
      ag_d <- ag_d +. (error -. ag_error);
      ag_error <- error;
      let accel = min 150.0 (p *. error +. i *. ag_i +. d *. ag_d) in
	(*Printf.printf "dist=%f accel=%f value=%f error=%f i=%f d=%f\n%!" (Vector.abs2 dest_delta) accel value error ag_i ag_d;*)
	self#accelerate accel

  method private autoguide (world:Object.world) (delta:float) =
    match goal with 
      | None -> ()
      | Some final_goal ->
	  let rec loop () =
	  match 
	    match path with 
	      | None -> 
		  let p = 
		    match PathPlanner.plan world at final_goal with
		      | None -> []
		      | Some p -> p
		  in
		    path <- Some p;
		    p
	      | Some p -> p
	  with
	    | [] when Vector.abs2 speed > 5.0 -> 
		self#autoguide_break delta
	    | [] -> goal <- None
	    | subgoal::rest ->  
		let dest_delta = subgoal -|. at in
		  if Vector.abs2 dest_delta < (if rest = [] then 10.0 else 20.0) then 
		    begin 
		      path <- Some rest;
		      loop ()
		    end
		  else
		    let goal_angle = Vector.angle dest_delta in
		    let angle_delta = 
		      let delta = GobaMath.angle_delta direction goal_angle in
			(* if we're about to turn more than 90 degrees while we're really near to the goal, don't *)
			if abs_float delta > pi /. 2.0 && Vector.abs2 (final_goal -|. at) < 50.0 then
			  GobaMath.angle_delta direction (goal_angle +. pi)
			else
			  delta
		    in
		    let turn = 
		      let max_angle = 90.0 /. 180.0 *. GobaMath.pi *. delta in
			abs_limit max_angle angle_delta
		    in
		      self#turn turn;
		      let (p, i, d) = (ag_k'p, ag_k'i, ag_k'd)
			(*	      match Vector.abs2 dest_delta with
			(*		| x when x < 50.0 -> (0.2, 0.1, 0.0)*)
				      | _ -> (0.0, 0.0, 0.5)*)
		      in
			(*let want_value = min 150.0 (Vector.dot dest_delta (cos direction, sin direction)) in
			  let value = Vector.dot speed (cos direction, sin direction) in*)
		      let want_value = 0.0 in
		      let value = 
			let v = ~-. (Vector.dot dest_delta (cos direction, sin direction)) in
			  v *. GobaMath.f_scale (pi /. 3.0) 0.0 (abs_float turn)
		      in
		      let error = GobaMath.abs_limit 0.2 (want_value -. value) in
			ag_i <- ag_i +. error;
			ag_d <- ag_d +. (error -. ag_error);
			ag_error <- error;
			let accel = min 150.0 (p *. error +. i *. ag_i +. d *. ag_d) in
			  (*Printf.printf "dist=%f accel=%f value=%f error=%f i=%f d=%f\n%!" (Vector.abs2 dest_delta) accel value error ag_i ag_d;*)
			  self#accelerate accel
	  in
	    loop ()

  method get_ammo = ammo

  method get_mines = mines

  method get_building_points = building_points

  method command (command:string) (args:string list) =
    let show () =  
      Printf.printf "k'p=%f k'i=%f k'd=%f\n%!" 
	ag_k'p
	ag_k'i
	ag_k'd
    in
      match command, args with
	| "s", _ -> 
	    show ()
	| "p", (value::_) -> 
	    ag_k'p <- (try float_of_string value with Failure _ -> ag_k'p);
	    show ()
	| "i", (value::_) -> 
	    ag_k'i <- (try float_of_string value with Failure _ -> ag_k'i);
	    show ()
	| "d", (value::_) -> 
	    ag_k'd <- (try float_of_string value with Failure _ -> ag_k'd);
	    show ()
	| _ -> ()

  method recharge =
    let ops = ref [] in
    let up c h = 
      let op () =
	if c () then begin h (); true; end else false 
      in
	ops := op :: !ops;
    in
      up 
	(fun () -> ammo < max_ammo)
	(fun () -> ammo <- ammo + 1);
      up
	(fun () -> hp < max_hp)
	(fun () -> hp <- hp +. 1.0);
      up
	(fun () -> mines < max_mines)
	(fun () -> mines <- mines + 1);
      up
	(fun () -> building_points < max_building_points)
	(fun () -> building_points <- building_points + 1);
      List.fold_left (fun recharged op -> op () || recharged) false !ops

  method signal_firing =
    if firing_time > firing_period && ammo > 0 then
      begin
	ammo <- ammo - 1;
	firing_time <- 0.0;
	self#change;
	true;
      end
    else
      false

  method signal_building n =
    if building_points >= n then
      begin
	building_points <- building_points - n;
	self#change;
	true
      end
    else
      false

  method signal_mine = 
    if mine_time > mine_period && mines > 0 then
      begin
	mine_time <- 0.0;
	mines <- mines - 1;
	self#change;
	true;
      end
    else
      false

  method collision_area = 
    Collision.carea_of_mask at (snd self#cur_frame)

  method cause_damage damage =
    if self#is_alive then
      begin
	hp <- max 0.0 (hp -. damage);
	damage_counter <- damage_counter +. sqrt damage /. 5.0;
	self#change;
      end

  method set_goal (at : float * float) =
    goal <- Some at;
    ag_i <- 0.0;
    ag_d <- 0.0;
    ag_error <- 0.0;
    path <- None;

  initializer
    (match init0 with 
       | Object.Init _ -> ()
       | Object.State (state0, age) -> self#set_state state0 age);
    self#change;
end
