TYPE_CONV_PATH "Goba"
module A = Bigarray.Array1

type contents = (char, Bigarray.int8_unsigned_elt, Bigarray.c_layout) A.t

let bs = 512

type header = 
    {
      name: string;			(* NUL-terminated if NUL fits *)
      mode: int;			(* *)
      uid: int;				(* *)
      gid: int;				(* *)
      size: int;			(* *)
      mtime: float;			(* *)
      chksum: int;			(* *)
      typeflag: char;			(* see below*)
      linkname: string;			(* NUL-terminated if NUL fits*)
      magic: string;			(* must be TMAGIC (NUL term.)*)
      version: string;			(* must be TVERSION*)
      uname: string;			(* NUL-terminated*)
      gname: string;			(* NUL-terminated*)
      devmajor: string;			(* *)
      devminor: string;			(* *)
      prefix: string;			(* NUL-terminated if NUL fits*)
   }

type file = 
    { header : header;
      data   : contents }

type t = { files : (string, file) Hashtbl.t }

let identity a = a

let int_of_oct s = 
  Scanf.sscanf s "%o" identity

let float_of_oct s = 
  Scanf.sscanf s "%Lo" Int64.to_float 

let string_of_contents a =
  let len = A.dim a in
  let s = String.create len in
    for c = 0 to len - 1 do
      s.[c] <- A.get a c
    done;
    s

let read_header : contents -> header option =
  fun contents ->
    try
      let r ofs len =
	(*Printf.printf "r %d %d (=%d)\n%!" ofs len (len + ofs);*)
	let s = string_of_contents (A.sub contents ofs len) in
	  try String.sub s 0 (String.index s '\000')
	  with Not_found -> s
      in
      let name = r 0 100 in
	if name <> "" then
	  Some
	    {
	      name = name;		(* NUL-terminated if NUL fits *)
	      mode = int_of_oct (r 100 8); (*  *)
	      uid = int_of_oct (r 108 8); (*  *)
	      gid = int_of_oct (r 116 8); (*  *)
	      size = int_of_oct (r 124 12); (*  *)
	      mtime = float_of_oct (r 136 12); (*  *)
	      chksum = int_of_oct (r 148 8); (*  *)
	      typeflag = ' ';		(* see below *)
	      linkname = r 157 100;	(* NUL-terminated if NUL fits *)
	      magic = r 257 6;		(* must be TMAGIC (NUL term.) *)
	      version = r 263 2;	(* must be TVERSION *)
	      uname = r 265 32;		(* NUL-terminated *)
	      gname = r 297 32;		(* NUL-terminated *)
	      devmajor = r 329 8;	(*  *)
	      devminor = r 337 8;	(*  *)
	      prefix = r 345 155;	(* NUL-terminated if NUL fits *)
	    }
	else
	  None
    with End_of_file ->  None


let scan_archive data =
  let length = A.dim data in
  let rec aux at files =
    if at < length then    
      let _ = () in
      match read_header (A.sub data at bs) with
	| Some header ->
	    let len = header.size in
	      aux 
		(at + bs + len + (512 - len mod bs) mod bs)
		({ header = header; data = A.sub data (at + bs) len }::files)
	| None -> List.rev files
    else
      List.rev files
  in
    aux 0 []

let hash_of_list f l =
  let h = Hashtbl.create (List.length l) in
    List.iter (fun l -> let (k, v) = f l in Hashtbl.add h k v) l;
    h

let open_archive file =
  let f = Unix.openfile file [Unix.O_RDONLY] 0 in
  let a = A.map_file f Bigarray.char Bigarray.c_layout false (-1) in
    Unix.close f;
    (* bring it to the memory to reduce random access.. *)
    let v = ref 0 in
      for c = 0 to A.dim a - 1 do v := !v + Char.code (A.get a c); done;
      { files = 
	  hash_of_list (fun ({ header = { name = name} } as file) -> (name, file)) 
	    (scan_archive a) }

let find t name = (Hashtbl.find t.files name).data

let contents t =
  List.rev (
    Hashtbl.fold
      (fun key _ files -> key::files)
      t.files
      []
  )
